#ifndef SAMPLEGRABBER_H
#define SAMPLEGRABBER_H

#include <vd2/system/thread.h>
#include <vd2/system/vdstl.h>
#include <vd2/system/refcount.h>
#include "BaseFilter.h"
#include "TransformPin.h"

class IDSSampleGrabberCallback {
public:
	virtual bool Process(IMediaSample *pSample, const REFERENCE_TIME& segmentStart) = 0;
	virtual void OnSampleMissed() = 0;
};

class IDSSampleGrabberTypeFilter : public IVDRefCount {
public:
	virtual bool IsMediaTypeSupported(const AM_MEDIA_TYPE *mt) = 0;
};

enum DSSampleGrabberCheckResult {
	kDSSampleGrabberCheckResult_NotFound,
	kDSSampleGrabberCheckResult_Partial,
	kDSSampleGrabberCheckResult_Completed
};

class DSSampleGrabberInputPin : public TransformInputPin {
public:
	DSSampleGrabberInputPin();
	~DSSampleGrabberInputPin();

	VDSignal& GetSignal() { return mSignal; }
	bool IsStreamEnded();

	void SetRunning(bool running);
	void SetMediaTypeFilter(IDSSampleGrabberTypeFilter *typeFilter) { mpTypeFilter = typeFilter; }

	REFERENCE_TIME GetCachedStartTime();
	REFERENCE_TIME GetCachedEndTime();
	DSSampleGrabberCheckResult CheckForSample(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend);
	void ResetOneShot(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend, bool forceAcceptLate);
	bool ResumeOneShot(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend);
	bool GetEstimatedStreamPosition(REFERENCE_TIME *pos);
	void SetCallback(IDSSampleGrabberCallback *pCB);
	void SetMediaType(const GUID *type);
	void SetMediaSubType(const GUID *subtype);

	HRESULT STDMETHODCALLTYPE QueryAccept(const AM_MEDIA_TYPE *pmt);
	HRESULT STDMETHODCALLTYPE Receive(IMediaSample *pSample);
	HRESULT STDMETHODCALLTYPE ReceiveCanBlock();
	HRESULT STDMETHODCALLTYPE BeginFlush();
	HRESULT STDMETHODCALLTYPE EndOfStream();
	HRESULT STDMETHODCALLTYPE NewSegment(REFERENCE_TIME tStart, REFERENCE_TIME tStop, double dRate);

	HRESULT STDMETHODCALLTYPE EnumMediaTypes(IEnumMediaTypes **ppEnum);

protected:
	void ClearQueueLocked();
	void ProcessPendingSampleLocked();

	GUID mMediaType;
	GUID mSubType;
	vdrefptr<IDSSampleGrabberTypeFilter> mpTypeFilter;

	// We use a distinct lock for shared data versus the streaming lock. This allows
	// us to hold the streaming lock to stop the stream while letting the processing
	// thread access the thread queues.
	VDCriticalSection mDataLock;
	REFERENCE_TIME mTargetTimeStart;
	REFERENCE_TIME mTargetTimeEnd;
	REFERENCE_TIME mTimeBase;
	REFERENCE_TIME mLastSeenTime;
	REFERENCE_TIME mLastSeenTimeSnapshot;
	bool mbRunning;
	bool mbStreamEnded;
	bool mbBufferSamples;
	bool mbSampleCaptured;
	bool mbForceAcceptLate;
	bool mbRestartOnNewSegment;
	IDSSampleGrabberCallback *mpCB;
	VDSignal mSignal;

	bool mResumeStreamingFlag;
	VDSignal mResumeStreamingSignal;

	struct SampleEntry {
		IMediaSample *mpSample;
		REFERENCE_TIME	mStartTime;
		REFERENCE_TIME	mEndTime;
	};

	typedef vdfastvector<SampleEntry> MediaSamples;
	MediaSamples mMediaSamples;

	SampleEntry mPendingSample;

	typedef vdfastvector<IMediaSample *> BufferCache;
	BufferCache mBufferCache;
};

class DSSampleGrabber : public BaseFilter {
public:
	DSSampleGrabber();
	~DSSampleGrabber();

	BasePin *GetInputPin() { return &mInputPin; }
	BasePin *GetOutputPin() { return &mOutputPin; }

	void SetCallback(IDSSampleGrabberCallback *pCB);
	void SetMediaType(const GUID *type);
	void SetMediaSubType(const GUID *subtype);
	void SetMediaTypeFilter(IDSSampleGrabberTypeFilter *typeFilter);
	bool GetEstimatedStreamPosition(REFERENCE_TIME *pos);
	REFERENCE_TIME GetCachedStartTime();
	REFERENCE_TIME GetCachedEndTime();
	DSSampleGrabberCheckResult CheckForSample(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend);
	void ResetOneShot(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend, bool forceAcceptLate);
	bool ResumeOneShot(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend);
	void WaitForSample();

	HRESULT STDMETHODCALLTYPE GetClassID(CLSID *pClassID);
	HRESULT STDMETHODCALLTYPE Stop();
	HRESULT STDMETHODCALLTYPE Pause();
	HRESULT STDMETHODCALLTYPE Run(REFERENCE_TIME tStart);

protected:
	DSSampleGrabberInputPin mInputPin;
	TransformOutputPin	mOutputPin;
};

#endif
