// DShowInputDriver: DirectShow based input plugin for VirtualDub
// Copyright (C) 2010-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#include "stdafx.h"
#include <vd2/system/math.h>
#include "AudioSourceDShow.h"
#include "Config.h"

namespace {
	// We start at 0.5 seconds and limit reseeking to 16 seconds.
	const sint64 kSeekFudgeOffsetMin = 5000000;
	const sint64 kSeekFudgeOffsetMax = 160000000;

	void AppendAudioFormatName(VDStringA& s, const AM_MEDIA_TYPE *amt) {
		if (amt->formattype == FORMAT_WaveFormatEx) {
			const WAVEFORMATEX& wfex = *(const WAVEFORMATEX *)amt->pbFormat;

			if (wfex.wFormatTag == WAVE_FORMAT_PCM) {
				s.append_sprintf("WFEX PCM %uHz, %u channels, %u bits, %u bytes/sec"
					, wfex.nSamplesPerSec
					, wfex.nChannels
					, wfex.wBitsPerSample
					, wfex.nAvgBytesPerSec
					);
			} else {
				s.append_sprintf("WFEX ID %02x, %uHz, %u channels, %u bytes/sec"
					, wfex.wFormatTag
					, wfex.nSamplesPerSec
					, wfex.nChannels
					, wfex.nAvgBytesPerSec
					);
			}
		} else {
			s += "Unknown format type";
		}
	}

	class StandardPCMMediaTypeFilter : public vdrefcounted<IDSSampleGrabberTypeFilter> {
	public:
		bool IsMediaTypeSupported(const AM_MEDIA_TYPE *mt) {
			if (mt->majortype != MEDIATYPE_Audio)
				return false;

			if (mt->formattype != FORMAT_WaveFormatEx)
				return false;

			if (mt->cbFormat < sizeof(WAVEFORMATEX) || !mt->pbFormat)
				return false;

			const WAVEFORMATEX *wfex = (const WAVEFORMATEX *)mt->pbFormat;

			if (wfex->wFormatTag != WAVE_FORMAT_PCM)
				return false;

			if (wfex->wBitsPerSample != 8 && wfex->wBitsPerSample != 16)
				return false;

			if (wfex->nChannels != 1 && wfex->nChannels != 2)
				return false;

			return true;
		}
	};
}

DSAudioSource::DSAudioSource(const VDXInputDriverContext *context, IDSEngine *engine, const wchar_t *fn)
	: mpEngine(engine)
	, mpContext(context)
	, mSeekFudgeOffset(0)
	, mRegisteredGraphId(0)
{
	memset(&mStreamInfo, 0, sizeof mStreamInfo);
	memset(&mAudioInfo, 0, sizeof mAudioInfo);

	try {
		DSEngineCommand cmd(&mCommandSignal);
		cmd.SetMethodCallback<DSAudioSource, &DSAudioSource::InitLocal>(this);
		cmd.SetExtraData((void *)fn);
		mpEngine->SendCommand(&cmd);
	} catch(const MyError&) {
		ShutdownLocal(NULL);
		throw;
	}
}

DSAudioSource::~DSAudioSource() {
	try {
		DSEngineCommand cmd(&mCommandSignal);
		cmd.SetMethodCallback<DSAudioSource, &DSAudioSource::ShutdownLocal>(this);
		mpEngine->SendCommand(&cmd);
	} catch(const MyError&) {
	}
}

void DSAudioSource::GetInfo(VDStringA& s) {
	s += mInfo;
}

// IVDXUnknown
int VDXAPIENTRY DSAudioSource::AddRef() {
	return vdxunknown<IVDXAudioSource>::AddRef();
}

int VDXAPIENTRY DSAudioSource::Release() {
	return vdxunknown<IVDXAudioSource>::Release();
}

void * VDXAPIENTRY DSAudioSource::AsInterface(uint32 iid) {
	if (iid == IVDXStreamSource::kIID)
		return static_cast<IVDXStreamSource *>(this);

	return vdxunknown<IVDXAudioSource>::AsInterface(iid);
}

// IVDXStreamSource
void VDXAPIENTRY DSAudioSource::GetStreamSourceInfo(VDXStreamSourceInfo& info) {
	info = mStreamInfo;
}

bool VDXAPIENTRY DSAudioSource::Read(sint64 lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {
	uint32 samples = lCount;
	uint32 bytes = lCount * mFormat->nBlockAlign;

	if (!lpBuffer) {
		*lBytesRead = bytes;
		*lSamplesRead = samples;
		return true;
	}

	if (bytes > cbBuffer) {
		uint32 bufferInSamples = cbBuffer / mFormat->nBlockAlign;

		if (!bufferInSamples) {
			*lBytesRead = mFormat->nBlockAlign;
			*lSamplesRead = 1;
			return false;
		}

		samples = bufferInSamples;
		bytes = samples * mFormat->nBlockAlign;
	}

	try {
		mReadPosition		= lStart;
		mReadSampleCount	= samples;
		mReadBytesLeft		= bytes;
		mReadBytesActual	= 0;
		mpReadBuffer		= lpBuffer;
		mbReadFirstSample	= true;

		DSEngineCommand cmd(&mCommandSignal);
		cmd.SetMethodCallback<DSAudioSource, &DSAudioSource::DecodeLocal>(this);
		cmd.SetExtraData(NULL);
		mpEngine->SendCommand(&cmd);
	} catch(const MyError& e) {
		mpContext->mpCallbacks->SetError("%s", e.gets());
		return false;
	}

	*lSamplesRead	= mReadBytesActual / mFormat->nBlockAlign;
	*lBytesRead		= mReadBytesActual - mReadBytesActual % mFormat->nBlockAlign;
	VDDEBUG("[AudioSource] Received %u samples\n", *lSamplesRead);
	return true;
}

const void *VDXAPIENTRY DSAudioSource::GetDirectFormat() {
	return mFormat.data();
}

int VDXAPIENTRY DSAudioSource::GetDirectFormatLen() {
	return mFormat.size();
}

IVDXStreamSource::ErrorMode VDXAPIENTRY DSAudioSource::GetDecodeErrorMode() {
	return IVDXStreamSource::kErrorModeReportAll;
}

void VDXAPIENTRY DSAudioSource::SetDecodeErrorMode(ErrorMode mode) {
}

bool VDXAPIENTRY DSAudioSource::IsDecodeErrorModeSupported(ErrorMode mode) {
	return mode == IVDXStreamSource::kErrorModeReportAll;
}

bool VDXAPIENTRY DSAudioSource::IsVBR() {
	return false;
}

sint64 VDXAPIENTRY DSAudioSource::TimeToPositionVBR(sint64 us) {
	return -1;
}

sint64 VDXAPIENTRY DSAudioSource::PositionToTimeVBR(sint64 samples) {
	return -1;
}

// IVDXAudioSource
void VDXAPIENTRY DSAudioSource::GetAudioSourceInfo(VDXAudioSourceInfo& info) {
	info = mAudioInfo;
}

void DSAudioSource::InitLocal(void *pvFn) {
	const wchar_t *fn = (const wchar_t *)pvFn;

	mInfo = "Audio stream:\n";

	HRESULT hr = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER, IID_IGraphBuilder, (void **)~mpGraph);
	if (FAILED(hr))
		throw MyError("Unable to create DirectShow filter graph: hr=%08x", hr);

	hr = mpGraph->QueryInterface(IID_IMediaControl, (void **)~mpGraphMediaControl);
	if (FAILED(hr))
		throw MyError("Unable to obtain DirectShow IMediaControl interface: hr=%08x", hr);

	hr = mpGraph->QueryInterface(IID_IMediaEvent, (void **)~mpGraphMediaEvent);
	if (FAILED(hr))
		throw MyError("Unable to obtain DirectShow IMediaEvent interface: hr=%08x", hr);

	hr = mpGraph->QueryInterface(IID_IMediaSeeking, (void **)~mpGraphMediaSeeking);
	if (FAILED(hr))
		throw MyError("Unable to obtain DirectShow IMediaSeeking interface: hr=%08x", hr);

	hr = mpGraph->QueryInterface(IID_IMediaFilter, (void **)~mpGraphMediaFilter);
	if (FAILED(hr))
		throw MyError("Unable to obtain DirectShow IMediaFilter interface: hr=%08x", hr);

	// kill the graph clock
	hr = mpGraphMediaFilter->SetSyncSource(NULL);
	if (FAILED(hr))
		throw MyError("Unable to disable DirectShow filter graph clock: hr=%08x", hr);

	mpSampleGrabber = new_nothrow DSSampleGrabber;
	if (!mpSampleGrabber)
		throw MyMemoryError();

	mpSampleGrabber->SetCallback(this);
	mpSampleGrabber->SetMediaType(&MEDIATYPE_Audio);

	DSConfigRegistryKey key;

	if (key.getBool("Audio: Force stereo", true))
		mpSampleGrabber->SetMediaTypeFilter(new StandardPCMMediaTypeFilter);

	hr = mpGraph->AddFilter(mpSampleGrabber, L"audio grabber");
	if (FAILED(hr))
		throw MyError("Unable to add sample grabber to graph: hr=%08x", hr);

	vdrefptr<ICaptureGraphBuilder2> pCB2;
	hr = CoCreateInstance(CLSID_CaptureGraphBuilder2, NULL, CLSCTX_INPROC_SERVER, IID_ICaptureGraphBuilder2, (void **)~pCB2);
	if (FAILED(hr))
		throw MyError("Unable to create capture graph builder: hr=%08x", hr);

	hr = pCB2->SetFiltergraph(mpGraph);
	if (FAILED(hr))
		throw MyError("Unable to set filter graph on capture graph builder: hr=%08x", hr);

	vdrefptr<IBaseFilter> pSourceFilter;
	hr = mpGraph->AddSourceFilter(fn, L"source", ~pSourceFilter);
	if (FAILED(hr))
		throw MyError("Unable to create source filter: hr=%08x", hr);

	hr = CoCreateInstance(CLSID_NullRenderer, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (void **)~mpRenderer);
	if (FAILED(hr))
		throw MyError("Unable to create null renderer: hr=%08x", hr);

	hr = mpGraph->AddFilter(mpRenderer, L"null renderer");
	if (FAILED(hr))
		throw MyError("Unable to add null renderer to graph: hr=%08x", hr);

	{
		vdrefptr<IPin> srcpin;
		hr = pCB2->FindPin(pSourceFilter, PINDIR_OUTPUT, NULL, &MEDIATYPE_Audio, TRUE, 0, ~srcpin);

		// We do two connection passes on either path, first with the two channel PCM filter,
		// and then again without it.
		if (FAILED(hr)) {
			hr = pCB2->RenderStream(NULL, NULL, pSourceFilter, mpSampleGrabber, mpRenderer);
			if (FAILED(hr)) {
				mpSampleGrabber->SetMediaTypeFilter(NULL);

				hr = pCB2->RenderStream(NULL, NULL, pSourceFilter, mpSampleGrabber, mpRenderer);
				if (FAILED(hr))
					throw MyError("Unable to render audio stream: hr=%08x", hr);
			}
		} else {
			hr = mpGraph->Connect(srcpin, mpSampleGrabber->GetInputPin());
			if (FAILED(hr)) {
				mpSampleGrabber->SetMediaTypeFilter(NULL);
				hr = mpGraph->Connect(srcpin, mpSampleGrabber->GetInputPin());

				if (FAILED(hr))
					throw MyError("Unable to render audio stream: hr=%08x", hr);
			}

			vdrefptr<IPin> rdrpin;
			hr = pCB2->FindPin(mpRenderer, PINDIR_INPUT, NULL, NULL, TRUE, 0, ~rdrpin);
			if (FAILED(hr))
				throw MyError("Unable to render audio stream: hr=%08x", hr);

			hr = mpGraph->Connect(mpSampleGrabber->GetOutputPin(), rdrpin);
			if (FAILED(hr))
				throw MyError("Unable to render audio stream: hr=%08x", hr);
		}
	}

	hr = mpSampleGrabber->GetInputPin()->ConnectedTo(~mpSourcePin);
	if (FAILED(hr))
		throw MyError("Unable to find source/decoder output pin: hr=%08x", hr);

	hr = mpSampleGrabber->GetOutputPin()->ConnectedTo(~mpRendererPin);
	if (FAILED(hr))
		throw MyError("Unable to find renderer input pin: hr=%08x", hr);

	// must get duration before we disconnect the source pin
	LONGLONG dur;
	hr = mpGraphMediaSeeking->GetDuration(&dur);
	if (FAILED(hr))
		throw MyError("Unable to get stream duration: hr=%08x", hr);

	mInfo.append_sprintf("  Duration\t%.7f seconds\n", (double)dur / 10000000.0);

	hr = mpSourcePin->ConnectedTo(~vdrefptr<IPin>());
	VDASSERT(SUCCEEDED(hr));

	hr = mpSampleGrabber->GetInputPin()->ConnectedTo(~vdrefptr<IPin>());
	VDASSERT(SUCCEEDED(hr));

	hr = mpSampleGrabber->GetOutputPin()->ConnectedTo(~vdrefptr<IPin>());
	VDASSERT(SUCCEEDED(hr));

	AM_MEDIA_TYPE connectedMediaType = {0};
	hr = mpSampleGrabber->GetInputPin()->ConnectionMediaType(&connectedMediaType);

	if (connectedMediaType.pUnk)
		connectedMediaType.pUnk->Release();

	if (FAILED(hr) || connectedMediaType.formattype != FORMAT_WaveFormatEx) {
		if (connectedMediaType.pbFormat)
			CoTaskMemFree(connectedMediaType.pbFormat);

		throw MyError("Could not find compatible audio decompression format.");
	}

	mInfo += "  Selected format\t";
	AppendAudioFormatName(mInfo, &connectedMediaType);
	mInfo += '\n';

	const WAVEFORMATEX& wfex = *(const WAVEFORMATEX *)connectedMediaType.pbFormat;

	mFormat.assign(&wfex, connectedMediaType.cbFormat);

	if (connectedMediaType.pbFormat)
		CoTaskMemFree(connectedMediaType.pbFormat);

	mStreamInfo.mSampleRate.mNumerator		= mFormat->nSamplesPerSec;
	mStreamInfo.mSampleRate.mDenominator	= 1;
	mStreamInfo.mPixelAspectRatio.mNumerator	= 0;
	mStreamInfo.mPixelAspectRatio.mDenominator	= 0;
	mStreamInfo.mSampleCount = VDRoundToInt64((double)mFormat->nSamplesPerSec * (double)dur / 10000000.0);

	// enumerate available formats from the audio decoder (diagnostic only)
	vdrefptr<IEnumMediaTypes> pMediaTypes;
	hr = mpSourcePin->EnumMediaTypes(~pMediaTypes);
	if (SUCCEEDED(hr)) {
		AM_MEDIA_TYPE *amt;
		ULONG fetched;

		while(S_OK == pMediaTypes->Next(1, &amt, &fetched)) {
			MediaTypePtr mt;
			mt.Attach(amt);

			if (mt->majortype == MEDIATYPE_Audio) {
				mInfo += "  Available format\t";
				AppendAudioFormatName(mInfo, mt);
				mInfo += '\n';
			}
		}
	}

	// add audio graph to ROT for GraphEdit
	DWORD id = 0;
	if (DSAddToRot(mpGraph, &id))
		mRegisteredGraphId = id;

	// add filter list to info string
	DSListFilters(mpGraph, mInfo);
}

void DSAudioSource::ShutdownLocal(void *) {
	if (mpGraphMediaControl) {
		mpGraphMediaControl->Stop();

		OAFilterState fs;
		mpGraphMediaControl->GetState(INFINITE, &fs);
	}

	if (mRegisteredGraphId) {
		DSRemoveFromRot(mRegisteredGraphId);
		mRegisteredGraphId = 0;
	}

	mpRendererPin = NULL;
	mpRenderer = NULL;
	mpSampleGrabber = NULL;
	mpSourcePin = NULL;
	mpGraphMediaSeeking = NULL;
	mpGraphMediaEvent = NULL;
	mpGraphMediaControl = NULL;
	mpGraphMediaFilter = NULL;
	mpGraph = NULL;
}

void DSAudioSource::DecodeLocal(void *) {
	// change time on sample grabber and see if it can grab the frame from its queue
	LONGLONG timeStartFlush = VDRoundToInt64((double)mReadPosition * (double)mStreamInfo.mSampleRate.mDenominator / (double)mStreamInfo.mSampleRate.mNumerator * 10000000.0);
	LONGLONG timeStart = VDRoundToInt64(((double)mReadPosition + 0.5) * (double)mStreamInfo.mSampleRate.mDenominator / (double)mStreamInfo.mSampleRate.mNumerator * 10000000.0);
	LONGLONG timeEnd = VDRoundToInt64(((double)(mReadPosition + mReadSampleCount) + 0.5) * (double)mStreamInfo.mSampleRate.mDenominator / (double)mStreamInfo.mSampleRate.mNumerator * 10000000.0);

	VDDEBUG("[AudioSource] Requesting %u samples (%u bytes) starting at %llu; setting search bracket to [%llu, %llu]; cached [%llu, %llu]\n"
		, mReadSampleCount
		, mReadBytesLeft
		, mReadPosition, timeStart, timeEnd
		, mpSampleGrabber->GetCachedStartTime()
		, mpSampleGrabber->GetCachedEndTime()
		);

	bool forceSeek = false;
	for(;;) {
		mReadTimeStart = timeStartFlush;
		mbSampleLate = false;

		DSSampleGrabberCheckResult checkResult = mpSampleGrabber->CheckForSample(timeStart, timeEnd);

		if (checkResult == kDSSampleGrabberCheckResult_Completed || (forceSeek && checkResult == kDSSampleGrabberCheckResult_NotFound)) {
			// We found all of the samples in the cache -- just exit.
			return;
		}

		// issue a seek
		HRESULT hr;
		OAFilterState fs;

		REFERENCE_TIME lastSeekPos;
		if (checkResult != kDSSampleGrabberCheckResult_Partial && (forceSeek || !mpSampleGrabber->GetEstimatedStreamPosition(&lastSeekPos) || timeStart < lastSeekPos || timeStart > lastSeekPos + 5000000)) {
			LONGLONG startPos = timeStartFlush - mSeekFudgeOffset;
			if (startPos < 0)
				startPos = 0;

			VDDEBUG("[audio] Issuing reseek to position %.7f\n", (double)startPos / 10000000.0);

			// lock down the sample grabber
			mpSampleGrabber->ResetOneShot(timeStart, timeEnd, mSeekFudgeOffset >= kSeekFudgeOffsetMax);

			hr = mpGraphMediaSeeking->SetPositions(&startPos, AM_SEEKING_AbsolutePositioning, NULL, AM_SEEKING_NoPositioning);
			if (FAILED(hr))
				throw MyError("Unable to seek DirectShow graph: error code %08x", hr);
		} else {
			if (mpSampleGrabber->ResumeOneShot(timeStart, timeEnd))
				return;
		}

		// run the graph
		hr = mpGraphMediaControl->GetState(INFINITE, &fs);
		if (FAILED(hr))
			throw MyError("Unable to run DirectShow graph: error code %08x", hr);

		if (fs != State_Running) {
			hr = mpGraphMediaControl->Run();
			if (FAILED(hr))
				throw MyError("Unable to run DirectShow graph: error code %08x", hr);

			// wait for state transition to complete
			hr = mpGraphMediaControl->GetState(INFINITE, &fs);
			if (FAILED(hr))
				throw MyError("Unable to run DirectShow graph: error code %08x", hr);
		}

		// wait for sample
		mpSampleGrabber->WaitForSample();

		if (!mbSampleLate)
			break;

		if (!mSeekFudgeOffset)
			mSeekFudgeOffset = kSeekFudgeOffsetMin;

		if (mSeekFudgeOffset < kSeekFudgeOffsetMax)
			mSeekFudgeOffset += mSeekFudgeOffset;

		forceSeek = true;
	}

	mpReadBuffer = NULL;
	mReadBytesLeft = 0;
}

bool DSAudioSource::Process(IMediaSample *pSample, const REFERENCE_TIME& timeBase) {
	REFERENCE_TIME t1, t2;
	HRESULT hr = pSample->GetTime(&t1, &t2);
	VDDEBUG("[AudioSource] Receiving audio: [%llu, %llu]\n", t1 + timeBase, t2 + timeBase);

	REFERENCE_TIME trelstart = mReadTimeStart - timeBase;
	int offsetInBytes = 0;
	if (mbReadFirstSample && SUCCEEDED(hr) && t1 < trelstart) {
		REFERENCE_TIME tdelta = trelstart - t1;
		int samples = VDRoundToInt((double)tdelta / (double)mStreamInfo.mSampleRate.mDenominator * (double)mStreamInfo.mSampleRate.mNumerator / 10000000.0);
		int bytes = samples * mFormat->nBlockAlign - mReadBytesActual;

		offsetInBytes = bytes;
	}
	mbReadFirstSample = false;

	BYTE *p;
	hr = pSample->GetPointer(&p);

	if (FAILED(hr)) {
		VDDEBUG("[AudioSource] GetPointer() call failed on media sample\n");
		return false;
	}

	long size = pSample->GetActualDataLength();
	VDASSERT(size >= 0);
	if (size <= 0) {
		VDDEBUG("[AudioSource] Rejecting media sample as it is empty\n");
		return true;
	}

	if (offsetInBytes < 0) {
		VDDEBUG("[AudioSource] Timestamps are advancing too quickly -- fudging offset from %d to 0.\n", offsetInBytes);
		offsetInBytes = 0;
	}

	if (offsetInBytes >= size) {
		// We used to assert that we were exactly on the end here, but we can't do that since
		// wonky timestamps can cause us to go slightly over.
		VDDEBUG("[AudioSource] Rejecting media sample as it is completely earlier than search range\n");
		return true;
	}

	uint32 usize = size - offsetInBytes;
	if (usize > mReadBytesLeft)
		usize = mReadBytesLeft;

	VDDEBUG("[AudioSource] Copying out %u bytes from offset %d\n", usize, offsetInBytes);

	memcpy(mpReadBuffer, p + offsetInBytes, usize);
	mpReadBuffer = (char *)mpReadBuffer + usize;
	mReadBytesLeft -= usize;
	mReadBytesActual += usize;

	return mReadBytesLeft > 0;
}

void DSAudioSource::OnSampleMissed() {
	mbSampleLate = true;
}
