// DShowInputDriver: DirectShow based input plugin for VirtualDub
// Copyright (C) 2010-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#include "stdafx.h"
#include "BaseFilter.h"

///////////////////////////////////////////////////////////////////////////////

class BaseFilterPinEnum : public IEnumPins {
	BaseFilterPinEnum& operator=(const BaseFilterPinEnum&);
public:
	BaseFilterPinEnum(BaseFilter *p);
	BaseFilterPinEnum(const BaseFilterPinEnum& src);
	~BaseFilterPinEnum();

	ULONG STDMETHODCALLTYPE AddRef();
	ULONG STDMETHODCALLTYPE Release();
	HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void **ppvObject);

	HRESULT STDMETHODCALLTYPE Next(ULONG cPins, IPin **ppPins, ULONG *pcFetched);
	HRESULT STDMETHODCALLTYPE Skip(ULONG cPins);
	HRESULT STDMETHODCALLTYPE Reset();
	HRESULT STDMETHODCALLTYPE Clone(IEnumPins **ppEnum);

protected:
	BaseFilter *mpBaseFilter;
	int mPinSetGeneration;
	size_t mPinIndex;
	LONG mRefCount;
};

BaseFilterPinEnum::BaseFilterPinEnum(BaseFilter *p)
	: mpBaseFilter(p)
	, mRefCount(0)
{
	mpBaseFilter->AddRef();

	Reset();

	mpBaseFilter->mLock.Enter();
	mPinIndex = 0;
	mPinSetGeneration = mpBaseFilter->mPinSetGeneration;
	mpBaseFilter->mLock.Exit();
}

BaseFilterPinEnum::BaseFilterPinEnum(const BaseFilterPinEnum& src)
	: mpBaseFilter(src.mpBaseFilter)
	, mPinSetGeneration(src.mPinSetGeneration)
	, mPinIndex(src.mPinIndex)
	, mRefCount(0)
{
}

BaseFilterPinEnum::~BaseFilterPinEnum() {
	mpBaseFilter->Release();
}

ULONG STDMETHODCALLTYPE BaseFilterPinEnum::AddRef() {
	return (ULONG)InterlockedIncrement(&mRefCount);
}

ULONG STDMETHODCALLTYPE BaseFilterPinEnum::Release() {
	LONG rv = InterlockedDecrement(&mRefCount);
	if (!rv)
		delete this;

	return rv;
}

HRESULT STDMETHODCALLTYPE BaseFilterPinEnum::QueryInterface(REFIID riid, void **ppvObject) {
	if (riid == IID_IEnumPins) {
		*ppvObject = static_cast<IEnumPins *>(this);
		AddRef();
		return S_OK;
	}

	if (riid == IID_IUnknown) {
		*ppvObject = static_cast<IUnknown *>(this);
		AddRef();
		return S_OK;
	}

	*ppvObject = NULL;
	return E_NOINTERFACE;
}

HRESULT STDMETHODCALLTYPE BaseFilterPinEnum::Next(ULONG cPins, IPin **ppPins, ULONG *pcFetched) {
	if (!ppPins)
		return E_POINTER;

	HRESULT hr;
	ULONG fetched = 0;

	mpBaseFilter->mLock.Enter();
	if (mPinSetGeneration != mpBaseFilter->mPinSetGeneration) {
		hr = VFW_E_ENUM_OUT_OF_SYNC;
	} else {
		size_t n = mpBaseFilter->mPinSet.size();
		ULONG pinsLeft = cPins;

		while(pinsLeft > 0 && mPinIndex < n) {
			BasePin *pin = mpBaseFilter->mPinSet[mPinIndex++];
			pin->AddRef();
			*ppPins++ = pin;
			--pinsLeft;
			++fetched;
		}

		if (pinsLeft)
			hr = S_FALSE;
		else
			hr = S_OK;
	}
	mpBaseFilter->mLock.Exit();

	if (pcFetched)
		*pcFetched = fetched;

	return hr;
}

HRESULT STDMETHODCALLTYPE BaseFilterPinEnum::Skip(ULONG cPins) {
	HRESULT hr;

	mpBaseFilter->mLock.Enter();
	if (mPinSetGeneration != mpBaseFilter->mPinSetGeneration)
		hr = VFW_E_ENUM_OUT_OF_SYNC;
	else {
		mPinIndex += cPins;
		if (mPinIndex >= mpBaseFilter->mPinSet.size())
			hr = S_FALSE;
		else
			hr = S_OK;
	}
	mpBaseFilter->mLock.Exit();

	return hr;
}

HRESULT STDMETHODCALLTYPE BaseFilterPinEnum::Reset() {
	mpBaseFilter->mLock.Enter();
	mPinIndex = 0;
	mPinSetGeneration = mpBaseFilter->mPinSetGeneration;
	mpBaseFilter->mLock.Exit();

	return S_OK;
}

HRESULT STDMETHODCALLTYPE BaseFilterPinEnum::Clone(IEnumPins **ppEnum) {
	if (!ppEnum)
		return E_POINTER;

	*ppEnum = new(std::nothrow) BaseFilterPinEnum(*this);
	if (!*ppEnum)
		return E_OUTOFMEMORY;

	return S_OK;
}

///////////////////////////////////////////////////////////////////////////////

BaseFilter::BaseFilter()
	: mpGraph(NULL)
	, mpNameInGraph(NULL)
	, mRefCount(0)
	, mPinSetGeneration(0)
	, mpClock(NULL)
	, mState(State_Stopped)
{
}

BaseFilter::~BaseFilter() {
	if (mpNameInGraph)
		SysFreeString(mpNameInGraph);
}

void BaseFilter::ReservePins(unsigned pins) {
	mLock.Enter();
	mPinSet.reserve(pins);
	mLock.Exit();
}

void BaseFilter::AddPin(BasePin *pin, LPCWSTR id) {
	pin->SetPinInfo(this, id);

	mLock.Enter();
	mPinSet.push_back(pin);
	++mPinSetGeneration;
	mLock.Exit();
}

bool BaseFilter::IsStopped() {
	mLock.Enter();
	bool stopped = (mState == State_Stopped);
	mLock.Exit();

	return stopped;
}

ULONG STDMETHODCALLTYPE BaseFilter::AddRef() {
	return (ULONG)InterlockedIncrement(&mRefCount);
}

ULONG STDMETHODCALLTYPE BaseFilter::Release() {
	LONG rc = InterlockedDecrement(&mRefCount);
	if (!rc)
		delete this;

	return (ULONG)rc;
}

HRESULT STDMETHODCALLTYPE BaseFilter::QueryInterface(REFIID riid, void **ppvObject) {
	if (riid == IID_IBaseFilter) {
		*ppvObject = static_cast<IBaseFilter *>(this);
		AddRef();
		return S_OK;
	}

	if (riid == IID_IMediaFilter) {
		*ppvObject = static_cast<IMediaFilter *>(this);
		AddRef();
		return S_OK;
	}

	if (riid == IID_IPersist) {
		*ppvObject = static_cast<IPersist *>(this);
		AddRef();
		return S_OK;
	}

	if (riid == IID_IUnknown) {
		*ppvObject = static_cast<IUnknown *>(this);
		AddRef();
		return S_OK;
	}

	*ppvObject = NULL;
	return E_NOINTERFACE;
}

HRESULT STDMETHODCALLTYPE BaseFilter::EnumPins(IEnumPins **ppEnum) {
	BaseFilterPinEnum *pPinEnum = new(std::nothrow) BaseFilterPinEnum(this);
	if (!pPinEnum)
		return E_OUTOFMEMORY;

	*ppEnum = pPinEnum;
	pPinEnum->AddRef();
	return S_OK;
}

HRESULT STDMETHODCALLTYPE BaseFilter::FindPin(LPCWSTR Id, IPin **ppPin) {
	HRESULT hr = VFW_E_NOT_FOUND;

	if (!Id || !ppPin)
		return E_POINTER;

	*ppPin = NULL;

	mLock.Enter();
	for(PinSet::const_iterator it(mPinSet.begin()), itEnd(mPinSet.end()); it != itEnd; ++it) {
		BasePin *pin = *it;

		if (!wcscmp(pin->GetId(), Id)) {
			*ppPin = pin;
			pin->AddRef();
			hr = S_OK;
		}
	}
	mLock.Exit();

	return hr;
}

HRESULT STDMETHODCALLTYPE BaseFilter::QueryFilterInfo(FILTER_INFO *pInfo) {
	if (mpNameInGraph) {
		wcsncpy(pInfo->achName, mpNameInGraph, MAX_FILTER_NAME - 1);
		pInfo->achName[MAX_FILTER_NAME - 1] = 0;
	} else {
		pInfo->achName[0] = 0;
	}

	pInfo->pGraph = mpGraph;
	if (pInfo->pGraph)
		pInfo->pGraph->AddRef();

	return S_OK;
}

HRESULT STDMETHODCALLTYPE BaseFilter::JoinFilterGraph(IFilterGraph *pGraph, LPCWSTR pName) {
	BSTR pNewName = NULL;
	
	if (pName) {
		pNewName = SysAllocString(pName);
		if (!pNewName)
			return E_OUTOFMEMORY;
	}

	mpGraph = pGraph;

	if (mpNameInGraph)
		SysFreeString(mpNameInGraph);

	mpNameInGraph = pNewName;

	return S_OK;
}

HRESULT STDMETHODCALLTYPE BaseFilter::QueryVendorInfo(LPWSTR *pVendorInfo) {
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE BaseFilter::Stop() {
	mLock.Enter();
	if (mState != State_Stopped) {
		for(PinSet::const_iterator it(mPinSet.begin()), itEnd(mPinSet.end()); it!=itEnd; ++it) {
			BasePin *pin = *it;
			pin->Deactivate();
		}
	}

	mState = State_Stopped;
	mLock.Exit();

	return S_OK;
}

HRESULT STDMETHODCALLTYPE BaseFilter::Pause() {
	HRESULT hr = S_OK;

	mLock.Enter();

	if (mState == State_Stopped) {
		for(PinSet::const_iterator it(mPinSet.begin()), itEnd(mPinSet.end()); it!=itEnd; ++it) {
			BasePin *pin = *it;
			hr = pin->Activate();
			if (hr != S_OK) {
				while (it != mPinSet.begin()) {
					pin = *it;
					pin->Deactivate();

					--it;
				}
				break;
			}
		}
	}

	if (hr == S_OK)
		mState = State_Paused;

	mLock.Exit();

	return hr;
}

HRESULT STDMETHODCALLTYPE BaseFilter::Run(REFERENCE_TIME tStart) {
	HRESULT hr = S_OK;

	mLock.Enter();

	if (mState == State_Stopped) {
		for(PinSet::const_iterator it(mPinSet.begin()), itEnd(mPinSet.end()); it!=itEnd; ++it) {
			BasePin *pin = *it;
			hr = pin->Activate();
			if (hr != S_OK) {
				while (it != mPinSet.begin()) {
					pin = *it;
					pin->Deactivate();

					--it;
				}
				break;
			}
		}
	}

	if (hr == S_OK)
		mState = State_Running;

	mLock.Exit();

	return hr;
}

HRESULT STDMETHODCALLTYPE BaseFilter::GetState(DWORD dwMilliSecsTimeout, FILTER_STATE *State) {
	if (!State)
		return E_POINTER;

	mLock.Enter();
	*State = mState;
	mLock.Exit();

	return S_OK;
}

HRESULT STDMETHODCALLTYPE BaseFilter::SetSyncSource(IReferenceClock *pClock) {
	if (pClock)
		pClock->AddRef();

	mLock.Enter();
	std::swap(mpClock, pClock);
	mLock.Exit();

	if (pClock)
		pClock->Release();

	return S_OK;
}

HRESULT STDMETHODCALLTYPE BaseFilter::GetSyncSource(IReferenceClock **pClock) {
	if (!pClock)
		return E_POINTER;

	mLock.Enter();
	IReferenceClock *pRefClock = mpClock;
	mLock.Exit();

	if (pRefClock)
		pRefClock->AddRef();

	*pClock = pRefClock;
	return S_OK;
}
