// DShowInputDriver: DirectShow based input plugin for VirtualDub
// Copyright (C) 2010-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#include "stdafx.h"
#include <vd2/system/math.h>
#include <vd2/system/fraction.h>
#include <vd2/system/VDString.h>
#include <vd2/system/registry.h>
#include <vd2/plugin/vdinputdriver.h>
#include <vd2/VDXFrame/Unknown.h>
#include "Config.h"
#include "Engine.h"
#include "BaseFilter.h"
#include "TransformPin.h"
#include "SampleGrabber.h"
#include "VideoSourceDShow.h"
#include "AudioSourceDShow.h"

#pragma comment(lib, "amstrmid")

void ShowUIDialogInfo(VDGUIHandle hwnd, const char *s);
void ShowUIDialogConfig(VDGUIHandle hwnd);
void ShowUIDialogAbout(VDGUIHandle hwnd);

///////////////////////////////////////////////////////////////////////////////

class DSInputFile : public vdxunknown<IVDXInputFile> {
public:
	DSInputFile(const VDXInputDriverContext *pContext, uint32 flags);
	~DSInputFile();

	int VDXAPIENTRY AddRef() { return vdxunknown<IVDXInputFile>::AddRef(); }
	int VDXAPIENTRY Release() { return vdxunknown<IVDXInputFile>::Release(); }
	void *VDXAPIENTRY AsInterface(uint32 iid) { return vdxunknown<IVDXInputFile>::AsInterface(iid); }

	bool VDXAPIENTRY PromptForOptions(VDXHWND, IVDXInputOptions **ppOptions);
	bool VDXAPIENTRY CreateOptions(const void *buf, uint32 len, IVDXInputOptions **ppOptions);

	void VDXAPIENTRY Init(const wchar_t *path, IVDXInputOptions *options);
	bool VDXAPIENTRY Append(const wchar_t *path);

	void VDXAPIENTRY DisplayInfo(VDXHWND hwndParent);

	bool VDXAPIENTRY GetVideoSource(int index, IVDXVideoSource **ppVS);
	bool VDXAPIENTRY GetAudioSource(int index, IVDXAudioSource **ppAS);

protected:
	void LocalInitFile(void *fnVoid);

	int mVideoStreams;
	int mAudioStreams;

	vdrefptr<IDSEngine> mpEngine;
	const VDXInputDriverContext *const mpContext;
	VDStringW mFilename;

	VDStringA mVideoInfo;
	VDStringA mAudioInfo;
};

DSInputFile::DSInputFile(const VDXInputDriverContext *pContext, uint32 flags)
	: mVideoStreams(0)
	, mAudioStreams(0)
	, mpContext(pContext)
{
}

DSInputFile::~DSInputFile() {
	if (mpEngine) {
		mpEngine->Shutdown();
		mpEngine = NULL;
	}
}

bool VDXAPIENTRY DSInputFile::PromptForOptions(VDXHWND, IVDXInputOptions **ppOptions) {
	return false;
}

bool VDXAPIENTRY DSInputFile::CreateOptions(const void *buf, uint32 len, IVDXInputOptions **ppOptions) {
	return false;
}

void VDXAPIENTRY DSInputFile::Init(const wchar_t *path, IVDXInputOptions *options) {
	try {
		mFilename = path;

		VDCreateDSEngine(~mpEngine);

		mpEngine->Init();

		VDSignal localSignal;
		DSEngineCommand cmd(&localSignal);
		cmd.SetMethodCallback<DSInputFile, &DSInputFile::LocalInitFile>(this);
		cmd.SetExtraData((void *)path);
		mpEngine->SendCommand(&cmd);
	} catch(const MyError& e) {
		mpContext->mpCallbacks->SetError("%s", e.gets());
	}
}

bool VDXAPIENTRY DSInputFile::Append(const wchar_t *path) {
	return false;
}

void VDXAPIENTRY DSInputFile::DisplayInfo(VDXHWND hwndParent) {
	VDStringA s;

	s = mVideoInfo;
	s += '\n';
	s += mAudioInfo;

	ShowUIDialogInfo((VDGUIHandle)hwndParent, s.c_str());
}

bool VDXAPIENTRY DSInputFile::GetVideoSource(int index, IVDXVideoSource **ppVS) {
	if (index || !mVideoStreams)
		return false;

	try {
		vdrefptr<DSVideoSource> vs(new DSVideoSource(mpContext, mpEngine, mFilename.c_str()));

		if (!index)
			vs->GetInfo(mVideoInfo);

		*ppVS = vs.release();
	} catch(const MyError& e) {
		mpContext->mpCallbacks->SetError("%s", e.gets());
		return false;
	}
	return true;
}

bool VDXAPIENTRY DSInputFile::GetAudioSource(int index, IVDXAudioSource **ppAS) {
	if (index || !mAudioStreams)
		return false;

	try {
		vdrefptr<DSAudioSource> as(new DSAudioSource(mpContext, mpEngine, mFilename.c_str()));

		if (!index)
			as->GetInfo(mAudioInfo);

		*ppAS = as.release();
	} catch(const MyError& e) {
		mpContext->mpCallbacks->SetError("%s", e.gets());
		return false;
	}
	return true;
}

void DSInputFile::LocalInitFile(void *fnVoid) {
	vdrefptr<IMediaDet> pDet;

	HRESULT hr = CoCreateInstance(CLSID_MediaDet, NULL, CLSCTX_INPROC_SERVER, IID_IMediaDet, (void **)~pDet);
	if (FAILED(hr))
		throw MyError("Unable to create media detector.");

	_bstr_t fn((const wchar_t *)fnVoid);

	hr = pDet->put_Filename(fn);
	if (FAILED(hr))
		throw MyError("Unable to set filename on media detector: hr = %08x", hr);

	long avStreams;
	hr = pDet->get_OutputStreams(&avStreams);
	if (FAILED(hr))
		throw MyError("Unable to get output stream count.");

	for(long i=0; i<avStreams; ++i) {
		hr = pDet->put_CurrentStream(i);
		if (FAILED(hr))
			throw MyError("Unable to peek at output stream %d: hr = %08x", i, hr);

		GUID majortype;
		hr = pDet->get_StreamType(&majortype);
		if (FAILED(hr))
			throw MyError("Unable to retrieve type of output stream %d: hr = %08x", i, hr);

		if (majortype == MEDIATYPE_Video)
			++mVideoStreams;
		else if (majortype == MEDIATYPE_Audio)
			++mAudioStreams;
	}

	// This is a hack to get some video files working that, for some reason, don't
	// detect properly with the media detector.
	if (mVideoStreams == 0)
		mVideoStreams = 1;
}

///////////////////////////////////////////////////////////////////////////////

class DSInputFileDriver : public vdxunknown<IVDXInputFileDriver> {
public:
	DSInputFileDriver(const VDXInputDriverContext *pContext);
	~DSInputFileDriver();

	int		VDXAPIENTRY DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize);
	bool	VDXAPIENTRY CreateInputFile(uint32 flags, IVDXInputFile **ppFile);

protected:
	const VDXInputDriverContext *const mpContext;
};

DSInputFileDriver::DSInputFileDriver(const VDXInputDriverContext *pContext)
	: mpContext(pContext)
{
}

DSInputFileDriver::~DSInputFileDriver() {
}

int	DSInputFileDriver::DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize) {
	return -1;
}

bool DSInputFileDriver::CreateInputFile(uint32 flags, IVDXInputFile **ppFile) {
	IVDXInputFile *p = new(std::nothrow) DSInputFile(mpContext, flags);
	if (!p)
		return false;

	*ppFile = p;
	p->AddRef();
	return true;
}

///////////////////////////////////////////////////////////////////////////////

bool VDXAPIENTRY DSCreateInputFileDriver(const VDXInputDriverContext *pContext, IVDXInputFileDriver **ppDriver) {
	IVDXInputFileDriver *p = new(std::nothrow) DSInputFileDriver(pContext);
	if (!p)
		return false;

	*ppDriver = p;
	p->AddRef();
	return true;
}

bool VDXAPIENTRY DSShowStaticConfigure(VDXHWND parent) {
	ShowUIDialogConfig((VDGUIHandle)parent);
	return true;
}

bool VDXAPIENTRY DSShowStaticAbout(VDXHWND parent) {
	ShowUIDialogAbout((VDGUIHandle)parent);
	return true;
}

VDStringW g_DSFileDetectPattern;
VDStringW g_DSFilePattern;
bool g_DSMatchRegistered;

VDXInputDriverDefinition g_inputDrv_DShow = {
	sizeof(VDXInputDriverDefinition),
	VDXInputDriverDefinition::kFlagSupportsVideo | VDXInputDriverDefinition::kFlagSupportsAudio | VDXInputDriverDefinition::kFlagCustomSignature,
	-10,
	NULL,
	0,
	NULL,					// filename pattern
	NULL,					// UI file pattern string
	L"DirectShow",
	DSCreateInputFileDriver
};

extern const VDXPluginInfo g_plugin_DShow = {
	sizeof(VDXPluginInfo),
	L"DirectShow",
	L"Avery Lee",
	L"DirectShow input driver",
	0,
	kVDXPluginType_Input,
	0,
	kVDXPlugin_APIVersion,
	kVDXPlugin_APIVersion,
	2,
	kVDXPlugin_InputDriverAPIVersion,
	&g_inputDrv_DShow,
	DSShowStaticAbout,
	DSShowStaticConfigure,
};

void DSInitPluginInfo() {
	if (!g_DSFilePattern.empty())
		return;

	g_DSFilePattern = L"DirectShow input driver|";

	DSConfigRegistryKey key;
	VDStringW extensions;
	if (key.getString("Extensions", extensions)) {
		VDStringRefW token;
		VDStringRefW xr(extensions);

		while(!xr.empty()) {
			if (!xr.split(';', token)) {
				token = xr;
				xr.clear();
			}

			if (!g_DSFileDetectPattern.empty())
				g_DSFileDetectPattern += '|';

			if (g_DSFilePattern.back() != '|')
				g_DSFilePattern += ';';

			g_DSFileDetectPattern.append(L"*.");
			g_DSFileDetectPattern.append(token.begin(), token.end());
			g_DSFilePattern.append(L"*.");
			g_DSFilePattern.append(token.begin(), token.end());
		}
	}

	g_DSMatchRegistered = key.getBool("Match registered", true);

	if (g_DSMatchRegistered) {
		HKEY hBaseKey;
		if (ERROR_SUCCESS == RegOpenKeyEx(HKEY_CLASSES_ROOT, "Media Type\\Extensions", 0, KEY_READ, &hBaseKey)) {
			char keyName[MAX_PATH];
			DWORD keyLen;

			for(DWORD dwIndex = 0; (keyLen = MAX_PATH), (RegEnumKeyEx(hBaseKey, dwIndex, keyName, &keyLen, NULL, NULL, NULL, NULL) == ERROR_SUCCESS); ++dwIndex) {
				if (!g_DSFileDetectPattern.empty())
					g_DSFileDetectPattern += '|';

				if (g_DSFilePattern.back() != '|')
					g_DSFilePattern += ';';

				g_DSFileDetectPattern.append_sprintf(L"*%hs", keyName);
				g_DSFilePattern.append_sprintf(L"*%hs", keyName);
			}

			RegCloseKey(hBaseKey);
		}
	}

	g_inputDrv_DShow.mpFilenameDetectPattern = g_DSFileDetectPattern.c_str();
	g_inputDrv_DShow.mpFilenamePattern = g_DSFilePattern.c_str();
}
