// DShowInputDriver: DirectShow based input plugin for VirtualDub
// Copyright (C) 2010-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#include "stdafx.h"
#include <vd2/system/vdtypes.h>
#include "MemPin.h"
#include "Allocator.h"

MemInputPin::MemInputPin()
	: mpAllocator(NULL)
{
}

MemInputPin::~MemInputPin() {
	if (mpAllocator)
		mpAllocator->Release();
}

void MemInputPin::Deactivate() {
	if (mpAllocator)
		mpAllocator->Decommit();
}

ULONG STDMETHODCALLTYPE MemInputPin::AddRef() {
	return InputPin::AddRef();
}

ULONG STDMETHODCALLTYPE MemInputPin::Release() {
	return InputPin::Release();
}

HRESULT STDMETHODCALLTYPE MemInputPin::QueryInterface(REFIID riid, void **ppvObject) {
	if (riid == IID_IMemInputPin) {
		*ppvObject = static_cast<IMemInputPin *>(this);
		AddRef();
		return S_OK;
	}

	return InputPin::QueryInterface(riid, ppvObject);
}

HRESULT STDMETHODCALLTYPE MemInputPin::GetAllocator(IMemAllocator **ppAllocator) {
	if (!ppAllocator)
		return E_POINTER;

	*ppAllocator = new(std::nothrow) MemAllocator;
	if (!*ppAllocator)
		return VFW_E_NO_ALLOCATOR;

	(*ppAllocator)->AddRef();
	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemInputPin::NotifyAllocator(IMemAllocator *pAllocator, BOOL bReadOnly) {
	if (!pAllocator)
		return E_POINTER;

	pAllocator->AddRef();
	if (mpAllocator)
		mpAllocator->Release();
	mpAllocator = pAllocator;

#if _DEBUG
	ALLOCATOR_PROPERTIES alprops = {};
	HRESULT hr = pAllocator->GetProperties(&alprops);
	if (SUCCEEDED(hr)) {
		VDDEBUG("NotifyAllocator: buffers=%u, size=%u, align=%u, prefix=%u\n", alprops.cBuffers, alprops.cbBuffer, alprops.cbAlign, alprops.cbPrefix);
	}
#endif

	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemInputPin::GetAllocatorRequirements(ALLOCATOR_PROPERTIES *pProps) {
	return E_NOTIMPL;
}

HRESULT STDMETHODCALLTYPE MemInputPin::Receive(IMediaSample *pSample) {
	return S_OK;
}

HRESULT STDMETHODCALLTYPE MemInputPin::ReceiveMultiple(IMediaSample **pSamples, long nSamples, long *nSamplesProcessed) {
	if (nSamples <= 0)
		return S_OK;

	if (!pSamples)
		return E_POINTER;

	long processed = 0;
	HRESULT hr = S_OK;
	for(long i=0; i<nSamples; ++i) {
		hr = Receive(*pSamples++);

		if (hr != S_OK)
			break;

		++processed;
	}

	if (nSamplesProcessed)
		*nSamplesProcessed = processed;

	return hr;
}

HRESULT STDMETHODCALLTYPE MemInputPin::ReceiveCanBlock() {
	return S_FALSE;
}

///////////////////////////////////////////////////////////////////////////////

MemOutputPin::MemOutputPin()
	: mpAllocator(NULL)
	, mpTransportPin(NULL)
{
}

MemOutputPin::~MemOutputPin() {
	if (mpAllocator) {
		mpAllocator->Release();
		mpAllocator = NULL;
	}

	if (mpTransportPin) {
		mpTransportPin->Release();
		mpTransportPin = NULL;
	}
}

HRESULT STDMETHODCALLTYPE MemOutputPin::Disconnect() {
	mLock.Enter();

	if (mpTransportPin) {
		mpTransportPin->Release();
		mpTransportPin = NULL;
	}

	HRESULT hr = OutputPin::Disconnect();
	mLock.Exit();

	return hr;
}

HRESULT STDMETHODCALLTYPE MemOutputPin::Connect(IPin *pReceivePin, const AM_MEDIA_TYPE *pmt) {
	if (!pReceivePin)
		return E_POINTER;

	mLock.Enter();

	if (mpConnectedPin) {
		mLock.Exit();
		return VFW_E_ALREADY_CONNECTED;
	}

	if (pmt) {
		HRESULT hr = pReceivePin->ReceiveConnection(this, pmt);
		if (FAILED(hr)) {
			mLock.Exit();
			return VFW_E_TYPE_NOT_ACCEPTED;
		}
	} else {
		IEnumMediaTypes *pEnum;

		HRESULT hr = EnumMediaTypes(&pEnum);
		if (FAILED(hr)) {
			mLock.Exit();
			return hr;
		}

		ULONG fetched;
		AM_MEDIA_TYPE *type;
		hr = E_FAIL;
		while(S_OK == pEnum->Next(1, &type, &fetched)) {
			hr = pReceivePin->ReceiveConnection(this, type);
			if (SUCCEEDED(hr))
				mpConnectedMediaType = type;

			if (type->pbFormat)
				CoTaskMemFree(type->pbFormat);
			CoTaskMemFree(type);

			if (SUCCEEDED(hr))
				break;
		}

		pEnum->Release();

		if (FAILED(hr)) {
			mLock.Exit();
			return VFW_E_NO_ACCEPTABLE_TYPES;
		}
	}

	IMemInputPin *pMemInputPin;
	HRESULT hr = pReceivePin->QueryInterface(IID_IMemInputPin, (void **)&pMemInputPin);
	if (FAILED(hr)) {
		mLock.Exit();
		return VFW_E_NO_TRANSPORT;
	}

	do {
		ALLOCATOR_PROPERTIES inreqs;
		ALLOCATOR_PROPERTIES reqs;

		reqs.cbBuffer	= 1;
		reqs.cbPrefix	= 0;
		reqs.cbAlign	= 4;
		reqs.cBuffers	= 1;

		hr = pMemInputPin->GetAllocatorRequirements(&inreqs);
		if (SUCCEEDED(hr)) {
			if (reqs.cbAlign < inreqs.cbAlign)
				reqs.cbAlign = inreqs.cbAlign;
			if (reqs.cbBuffer < inreqs.cbBuffer)
				reqs.cbBuffer = inreqs.cbBuffer;
			if (reqs.cbPrefix < inreqs.cbPrefix)
				reqs.cbPrefix = inreqs.cbPrefix;
			if (reqs.cBuffers < inreqs.cBuffers)
				reqs.cBuffers = inreqs.cBuffers;
		} else if (hr != E_NOTIMPL)
			break;

		// check if the input pin has an allocator
		IMemAllocator *pAllocator;
		hr = pMemInputPin->GetAllocator(&pAllocator);
		if (FAILED(hr)) {
			MemAllocator *pMemAllocator = new(std::nothrow) MemAllocator;
			if (!pMemAllocator) {
				hr = E_OUTOFMEMORY;
				break;
			}

			hr = pMemAllocator->Init();
			if (FAILED(hr))
				break;

			pAllocator = pMemAllocator;
		}

		ALLOCATOR_PROPERTIES outreqs;
		hr = pAllocator->SetProperties(&reqs, &outreqs);
		if (FAILED(hr)) {
			pAllocator->Release();
			break;
		}

		hr = pMemInputPin->NotifyAllocator(pAllocator, FALSE);
		if (FAILED(hr)) {
			pAllocator->Release();
			break;
		}

		if (mpAllocator)
			mpAllocator->Release();
		mpAllocator = pAllocator;

		hr = S_OK;
	} while(false);

	if (SUCCEEDED(hr)) {
		mpConnectedPin = pReceivePin;
		mpTransportPin = pMemInputPin;
		pReceivePin->AddRef();
	} else {
		pReceivePin->Disconnect();
		pMemInputPin->Release();
	}

	mLock.Exit();
	return hr;
}
