// DShowInputDriver: DirectShow based input plugin for VirtualDub
// Copyright (C) 2010-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#include "stdafx.h"
#include "Allocator.h"
#include "SampleGrabber.h"

#define VDDEBUG_DSSG VDDEBUG
//#define VDDEBUG_DSSG (void)sizeof printf

namespace {
	const REFERENCE_TIME kInfinityEndTime = 0x1fffffffffffffffull;	// leave some buffer
}

///////////////////////////////////////////////////////////////////////////////

DSSampleGrabberInputPin::DSSampleGrabberInputPin()
	: mMediaType(GUID_NULL)
	, mSubType(GUID_NULL)
	, mTargetTimeStart(0)
	, mTargetTimeEnd(0)
	, mTimeBase(0)
	, mLastSeenTime(0)
	, mLastSeenTimeSnapshot(0)
	, mpCB(NULL)
	, mbRunning(false)
	, mbStreamEnded(false)
	, mbBufferSamples(false)
	, mbSampleCaptured(false)
	, mbForceAcceptLate(false)
	, mbRestartOnNewSegment(false)
	, mResumeStreamingFlag(false)
{
	mPendingSample.mpSample = NULL;
}

DSSampleGrabberInputPin::~DSSampleGrabberInputPin() {
	while(!mBufferCache.empty()) {
		IMediaSample *p = mBufferCache.back();
		mBufferCache.pop_back();

		p->Release();
	}
}

bool DSSampleGrabberInputPin::GetEstimatedStreamPosition(REFERENCE_TIME *pos) {
	bool success = false;

	mDataLock.Lock();
	if (mbBufferSamples) {
		success = true;
		*pos = mLastSeenTimeSnapshot;
	}
	mDataLock.Unlock();

	return success;
}

bool DSSampleGrabberInputPin::IsStreamEnded() {
	mDataLock.Lock();
	const bool ended = mbStreamEnded;
	mDataLock.Unlock();

	return ended;
}

void DSSampleGrabberInputPin::SetRunning(bool running) {
	mDataLock.Lock();
	mbRunning = running;
	if (!running)
		ClearQueueLocked();
	mResumeStreamingFlag = true;
	mResumeStreamingSignal.signal();
	mDataLock.Unlock();
}

REFERENCE_TIME DSSampleGrabberInputPin::GetCachedStartTime() {
	mDataLock.Lock();
	REFERENCE_TIME startTime = mMediaSamples.empty() ? 0 : mMediaSamples.front().mStartTime + mTimeBase;
	mDataLock.Unlock();

	return startTime;
}

REFERENCE_TIME DSSampleGrabberInputPin::GetCachedEndTime() {
	mDataLock.Lock();
	REFERENCE_TIME endTime = mMediaSamples.empty() ? 0 : mMediaSamples.back().mEndTime + mTimeBase;
	mDataLock.Unlock();

	return endTime;
}

DSSampleGrabberCheckResult DSSampleGrabberInputPin::CheckForSample(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend) {
	// Check if we have a sample that contains this time -- if so, process it immediately.
	// Discard all samples that are too old.
	mDataLock.Lock();

	REFERENCE_TIME trelstart(tstart - mTimeBase);
	REFERENCE_TIME trelend(tend - mTimeBase);

	MediaSamples::iterator it(mMediaSamples.begin()), itEnd(mMediaSamples.end());
	DSSampleGrabberCheckResult result = kDSSampleGrabberCheckResult_NotFound;

	VDDEBUG_DSSG("SampleGrabber[%ls]: CheckForSample[%llu, %llu]\n", mpFilter->GetNameInGraph(), tstart, tend);
	while(it != itEnd) {
		const SampleEntry& samp = *it;

		// End time <= desired time? Discard.
		if (samp.mEndTime <= trelstart) {
			VDDEBUG_DSSG("SampleGrabber[%ls]: Discarding sample [%llu, %llu]\n", mpFilter->GetNameInGraph(), samp.mStartTime + mTimeBase, samp.mEndTime + mTimeBase);
			//samp.mpSample->Release();
			mBufferCache.push_back(samp.mpSample);

			it = mMediaSamples.erase(it);
			itEnd = mMediaSamples.end();
			continue;
		}

		// Contains our time? Take it.
		if (trelstart < samp.mEndTime && samp.mStartTime < trelend) {
			bool cont = true;

			VDDEBUG_DSSG("SampleGrabber[%ls]: Pushing sample [%llu, %llu] from buffer %p\n", mpFilter->GetNameInGraph(), samp.mStartTime + mTimeBase, samp.mEndTime + mTimeBase, samp.mpSample);

			if (mpCB)
				cont = mpCB->Process(samp.mpSample, mTimeBase);

			// We only return success if the callback says we are done. The reason is that it's possible
			// that we think we have a sample that covers the given time range, but when we actually check
			// the data length in Process() it turns out not to have the desired data. This can happen if
			// the decoder is a bit sloppy about timestamps and in particular pushes media samples with
			// ranges that sometimes don't abut properly.
			if (!cont) {
				mbSampleCaptured = true;
				result = kDSSampleGrabberCheckResult_Completed;
				break;
			}

			result = kDSSampleGrabberCheckResult_Partial;
		}

		++it;
	}

	VDDEBUG_DSSG("SampleGrabber[%ls]: CheckForSample() completed\n", mpFilter->GetNameInGraph());

	mLastSeenTimeSnapshot = mLastSeenTime;

	if (result == kDSSampleGrabberCheckResult_Partial && mbStreamEnded)
		result = kDSSampleGrabberCheckResult_Completed;

	mDataLock.Unlock();

	return result;
}

void DSSampleGrabberInputPin::ResetOneShot(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend, bool forceAcceptLate) {
	// grab the data lock and flush the queues
	mDataLock.Lock();
	mTargetTimeStart = tstart;
	mTargetTimeEnd = tend;
	mbBufferSamples = false;
	mbSampleCaptured = true;
	mbForceAcceptLate = forceAcceptLate;
	mbRestartOnNewSegment = true;
	mSignal.check();

	ClearQueueLocked();

	mResumeStreamingFlag = true;
	mResumeStreamingSignal.signal();
	mDataLock.Unlock();
}

bool DSSampleGrabberInputPin::ResumeOneShot(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend) {
	mDataLock.Lock();

	mbSampleCaptured = false;
	mbForceAcceptLate = false;
	mTargetTimeStart = tstart;
	mTargetTimeEnd = tend;
	mbBufferSamples = true;
	mSignal.check();

	mResumeStreamingFlag = true;
	mResumeStreamingSignal.signal();
	mDataLock.Unlock();

	return false;
}

void DSSampleGrabberInputPin::ClearQueueLocked() {
	while(!mMediaSamples.empty()) {
		IMediaSample *samp = mMediaSamples.back().mpSample;
		mMediaSamples.pop_back();
		mBufferCache.push_back(samp);
	}

	if (mPendingSample.mpSample) {
		mBufferCache.push_back(mPendingSample.mpSample);
		mPendingSample.mpSample = NULL;
	}
}

void DSSampleGrabberInputPin::ProcessPendingSampleLocked() {
	if (!mPendingSample.mpSample)
		return;

	vdrefptr<IMediaSample> sample;
	sample.set(mPendingSample.mpSample);
	mPendingSample.mpSample = NULL;

	// reject if sample time bracket is prior to desired time
	REFERENCE_TIME t1 = mPendingSample.mStartTime;
	REFERENCE_TIME t2 = mPendingSample.mEndTime;
	REFERENCE_TIME trelstart = mTargetTimeStart - mTimeBase;
	REFERENCE_TIME trelend = mTargetTimeEnd - mTimeBase;

	VDDEBUG_DSSG("SampleGrabber[%ls]: Processing pending sample [%llu, %llu]\n", mpFilter->GetNameInGraph(), t1 + mTimeBase, t2 + mTimeBase);

	if (t2 <= trelstart)
		return;

	if (t1 >= trelend && !mbSampleCaptured) {
		if (!mbForceAcceptLate) {
			VDDEBUG_DSSG("SampleGrabber[%ls]: Time missed -- aborting ([%llu,%llu] after [%llu,%llu])\n", mpFilter->GetNameInGraph(), t1 + mTimeBase, t2 + mTimeBase, mTargetTimeStart, mTargetTimeEnd);

			if (mpCB)
				mpCB->OnSampleMissed();

			mbSampleCaptured = true;
			mSignal.signal();
			return;
		}

		VDDEBUG_DSSG("SampleGrabber[%ls]: Time missed -- feeding late sample ([%d,%d] after [%d,%d])\n", mpFilter->GetNameInGraph(), (int)t1, (int)t2, (int)trelstart, (int)trelend);
		t1 = trelstart;
	}

	mLastSeenTime = t2 + mTimeBase;

	if (!mbSampleCaptured) {
		bool cont = true;

		if (mpCB)
			cont = mpCB->Process(sample, mTimeBase);

		if (!cont) {
			mSignal.signal();
			mbSampleCaptured = true;
		}
	}

	if (mbBufferSamples) {
		VDASSERT(mMediaSamples.empty() || mMediaSamples.back().mStartTime != t1);
		SampleEntry& ent = mMediaSamples.push_back();
		ent.mpSample = sample.release();

		ent.mStartTime = t1;
		ent.mEndTime = t2;
	}
}

void DSSampleGrabberInputPin::SetCallback(IDSSampleGrabberCallback *pCB) {
	mpCB = pCB;
}

void DSSampleGrabberInputPin::SetMediaType(const GUID *type) {
	mMediaType = *type;
}

void DSSampleGrabberInputPin::SetMediaSubType(const GUID *subtype) {
	mSubType = *subtype;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberInputPin::QueryAccept(const AM_MEDIA_TYPE *pmt) {
	if (mMediaType != GUID_NULL) {
		if (pmt->majortype != mMediaType)
			return S_FALSE;

		if (mSubType != GUID_NULL && pmt->subtype != mSubType)
			return S_FALSE;

		if (mMediaType == MEDIATYPE_Video) {
			if (pmt->formattype != FORMAT_VideoInfo && pmt->formattype != FORMAT_VideoInfo2)
				return S_FALSE;

			if (mSubType == GUID_NULL) {
				if (pmt->subtype != MEDIASUBTYPE_RGB24
					&& pmt->subtype != MEDIASUBTYPE_RGB32
					&& pmt->subtype != MEDIASUBTYPE_YUY2
					&& pmt->subtype != MEDIASUBTYPE_UYVY
					&& pmt->subtype != MEDIASUBTYPE_YV12
					&& pmt->subtype != MEDIASUBTYPE_YVU9
					&& pmt->subtype != MEDIASUBTYPE_RGB555
					&& pmt->subtype != MEDIASUBTYPE_RGB565)
					return S_FALSE;
			}
		} else if (mMediaType == MEDIATYPE_Audio) {
			if (pmt->formattype != FORMAT_WaveFormatEx)
				return S_FALSE;

			if (mSubType == GUID_NULL) {
				if (pmt->subtype != MEDIASUBTYPE_PCM
					&& pmt->subtype != MEDIASUBTYPE_PCMAudio_Obsolete)
					return S_FALSE;
			}
		}
	}

	if (mpTypeFilter && !mpTypeFilter->IsMediaTypeSupported(pmt))
		return S_FALSE;

	return TransformInputPin::QueryAccept(pmt);
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberInputPin::Receive(IMediaSample *pSample) {
	HRESULT hr = S_OK;

	// grab streaming lock
	mLock.Enter();

	// grab data lock
	mDataLock.Lock();

	// If we're flushing or not actively capturing, just ignore the sample.
restart:
	if (mbFlushing || !mbRunning) {
		hr = S_FALSE;
	} else if (mbBufferSamples || !mbSampleCaptured) {
		REFERENCE_TIME t1, t2;
		hr = pSample->GetTime(&t1, &t2);

		if (hr == S_OK) {
			// Check if we have too many buffers; block the queue if so.
			if (mMediaSamples.size() > 2) {
				if (mbSampleCaptured) {
					mResumeStreamingFlag = false;
				} else {
					mMediaSamples.front().mpSample->Release();
					mMediaSamples.erase(mMediaSamples.begin());
				}
			}

			// Wait for the signal to resume streaming. This MUST be done with the streaming lock held,
			// but NOT the data lock.
			if (!mResumeStreamingFlag) {
				do {
					mDataLock.Unlock();
					mResumeStreamingSignal.wait();
					mDataLock.Lock();
				} while(!mResumeStreamingFlag);

				// recheck flags
				goto restart;
			}

			long dataLength = pSample->GetActualDataLength();
			long bufferSize = pSample->GetSize();

			if (bufferSize < dataLength)
				bufferSize = dataLength;

			VDDEBUG_DSSG("SampleGrabber[%ls]: Receiving [%llu, %llu] size %d\n", mpFilter->GetNameInGraph(), t1 + mTimeBase, t2 + mTimeBase, dataLength);

			// end the last open sample
			if (mPendingSample.mpSample) {
				mPendingSample.mEndTime = t1;
				ProcessPendingSampleLocked();
			}

			// copy the sample
			vdrefptr<IMediaSample> pNewSample;

			while(!mBufferCache.empty()) {
				IMediaSample *ms = mBufferCache.back();
				mBufferCache.pop_back();

				// Check if the buffer's big enough; if not, just dump it. Ideally this should never happen
				// as we allocate according to the original sample's buffer size, not data length.
				long actualSize = ms->GetSize();
				if (actualSize >= dataLength) {
					pNewSample.set(ms);
					break;
				}

				// dump the buffer and try again
				ms->Release();
			}

			// allocate a new sample if we need one
			if (!pNewSample) {
				pNewSample = new_nothrow MemMediaSample;
				if (pNewSample && !static_cast<MemMediaSample *>(&*pNewSample)->Init(bufferSize))
					pNewSample = NULL;
			}

			if (pNewSample) {
				// copy data
				BYTE *src;
				hr = pSample->GetPointer(&src);
				if (SUCCEEDED(hr)) {
					BYTE *dst;
					pNewSample->GetPointer(&dst);	// never fails

					memcpy(dst, src, dataLength);

					// copy over most properties; we omit the media type, however
					pNewSample->SetTime(&t1, &t2);
					pNewSample->SetActualDataLength(dataLength);
					pNewSample->SetDiscontinuity(pSample->IsDiscontinuity() == S_OK);
					pNewSample->SetPreroll(pSample->IsPreroll() == S_OK);
					pNewSample->SetSyncPoint(pSample->IsSyncPoint() == S_OK);

					LONGLONG mediaStart, mediaEnd;
					HRESULT hr2 = pSample->GetMediaTime(&mediaStart, &mediaEnd);
					if (SUCCEEDED(hr2))
						pNewSample->SetMediaTime(&mediaStart, &mediaEnd);
					else
						pNewSample->SetMediaTime(NULL, NULL);

					// push a new pending sample
					VDASSERT(!mPendingSample.mpSample);
					mPendingSample.mpSample = pNewSample.release();
					mPendingSample.mStartTime = t1;
					mPendingSample.mEndTime = t1;
				}
			} else {
				hr = E_OUTOFMEMORY;
				mbSampleCaptured = true;
			}
		}
	}

	mDataLock.Unlock();

	if (mpAssocOutputPin)
		mpAssocOutputPin->Output(pSample);
	mLock.Exit();

	return hr;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberInputPin::ReceiveCanBlock() {
	return S_OK;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberInputPin::BeginFlush() {
	VDDEBUG_DSSG("SampleGrabber[%ls]: BeginFlush() called\n");

	mDataLock.Lock();
	ClearQueueLocked();
	mbBufferSamples = false;
	mResumeStreamingFlag = true;
	mbStreamEnded = true;
	mDataLock.Unlock();

	return TransformInputPin::BeginFlush();
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberInputPin::EndOfStream() {
	mDataLock.Lock();
	VDDEBUG_DSSG("SampleGrabber[%ls]: EndOfStream() called\n", mpFilter->GetNameInGraph());
	mbBufferSamples = false;
	mbStreamEnded = true;

	// If we have an open sample, set its end time to infinity and queue it
	if (mPendingSample.mpSample) {
		mPendingSample.mEndTime = kInfinityEndTime;

		ProcessPendingSampleLocked();
	}

	if (!mbSampleCaptured) {
		mbSampleCaptured = true;
		// It may seem to do this under lock, but it's so that CheckForSample() can atomically
		// clear the signal state while changing state variables.
		mSignal.signal();
	}
	mDataLock.Unlock();

	HRESULT hr = TransformInputPin::EndOfStream();
	return hr;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberInputPin::NewSegment(REFERENCE_TIME tStart, REFERENCE_TIME tStop, double dRate) {
	VDDEBUG_DSSG("SampleGrabber[%ls]: NewSegment(%lld, %lld, %g)\n", mpFilter->GetNameInGraph(), tStart, tStop, dRate);
	mDataLock.Lock();
	mTimeBase = tStart;
	HRESULT hr = TransformInputPin::NewSegment(tStart, tStop, dRate);
	ClearQueueLocked();

	if (mbRestartOnNewSegment) {
		VDDEBUG_DSSG("SampleGrabber[%ls]: NewSegment() called; restarting capture.\n", mpFilter->GetNameInGraph());
		mbRestartOnNewSegment = false;
		mbSampleCaptured = false;
		mbBufferSamples = true;
	}

	mbStreamEnded = false;

	mDataLock.Unlock();

	return hr;
}

class DSSampleGrabberMediaTypeEnumerator : public IEnumMediaTypes {
	DSSampleGrabberMediaTypeEnumerator& operator=(const DSSampleGrabberMediaTypeEnumerator&);
public:
	DSSampleGrabberMediaTypeEnumerator(const GUID& mediaType, const GUID& subType);
	DSSampleGrabberMediaTypeEnumerator(const DSSampleGrabberMediaTypeEnumerator&);
	~DSSampleGrabberMediaTypeEnumerator();

	ULONG STDMETHODCALLTYPE AddRef();
	ULONG STDMETHODCALLTYPE Release();
	HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void **ppvObject);

	HRESULT STDMETHODCALLTYPE Next(ULONG cMediaTypes, AM_MEDIA_TYPE **ppMediaTypes, ULONG *pcFetched);
	HRESULT STDMETHODCALLTYPE Skip(ULONG cMediaTypes);
	HRESULT STDMETHODCALLTYPE Reset();
	HRESULT STDMETHODCALLTYPE Clone(IEnumMediaTypes **ppEnum);

protected:
	LONG mRefCount;
	GUID mMediaType;
	GUID mSubType;
	int mIndex;
};

DSSampleGrabberMediaTypeEnumerator::DSSampleGrabberMediaTypeEnumerator(const GUID& mediaType, const GUID& subType)
	: mRefCount(0)
	, mMediaType(mediaType)
	, mSubType(subType)
	, mIndex(0)
{
}

DSSampleGrabberMediaTypeEnumerator::DSSampleGrabberMediaTypeEnumerator(const DSSampleGrabberMediaTypeEnumerator& src)
	: mRefCount(0)
	, mMediaType(src.mMediaType)
	, mSubType(src.mSubType)
	, mIndex(src.mIndex)
{
}

DSSampleGrabberMediaTypeEnumerator::~DSSampleGrabberMediaTypeEnumerator() {
}

ULONG STDMETHODCALLTYPE DSSampleGrabberMediaTypeEnumerator::AddRef() {
	return (ULONG)InterlockedIncrement(&mRefCount);
}

ULONG STDMETHODCALLTYPE DSSampleGrabberMediaTypeEnumerator::Release() {
	LONG rv = InterlockedDecrement(&mRefCount);

	if (!rv)
		delete this;

	return (ULONG)rv;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberMediaTypeEnumerator::QueryInterface(REFIID riid, void **ppvObject) {
	if (riid == IID_IEnumMediaTypes) {
		*ppvObject = static_cast<IEnumMediaTypes *>(this);
		AddRef();
		return S_OK;
	}

	if (riid == IID_IUnknown) {
		*ppvObject = static_cast<IUnknown *>(this);
		AddRef();
		return S_OK;
	}

	*ppvObject = NULL;
	return E_NOINTERFACE;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberMediaTypeEnumerator::Next(ULONG cMediaTypes, AM_MEDIA_TYPE **ppMediaTypes, ULONG *pcFetched) {
	if (cMediaTypes && !ppMediaTypes)
		return E_POINTER;

	if (!cMediaTypes) {
		if (pcFetched)
			*pcFetched = 0;

		return S_OK;
	}

	ULONG left = 1 - mIndex;

	if (left) {
		AM_MEDIA_TYPE t = {};
		t.majortype = mMediaType;
		t.subtype = mSubType;
		t.formattype = FORMAT_None;

		MediaTypePtr p(&t);

		ppMediaTypes[0] = p;

		if (!p)
			return E_OUTOFMEMORY;

		p.Detach();
	}

	if (pcFetched)
		*pcFetched = left;

	mIndex = 1;

	return left ? S_OK : S_FALSE;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberMediaTypeEnumerator::Skip(ULONG cMediaTypes) {
	if (!cMediaTypes)
		return S_OK;

	if (mIndex == 0 && cMediaTypes == 1) {
		mIndex = 1;
		return S_OK;
	}

	mIndex = 1;
	return S_FALSE;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberMediaTypeEnumerator::Reset() {
	mIndex = 0;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberMediaTypeEnumerator::Clone(IEnumMediaTypes **ppEnum) {
	if (!ppEnum)
		return E_POINTER;

	*ppEnum = new(std::nothrow) DSSampleGrabberMediaTypeEnumerator(*this);
	if (!*ppEnum)
		return E_OUTOFMEMORY;

	(*ppEnum)->AddRef();
	return S_OK;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabberInputPin::EnumMediaTypes(IEnumMediaTypes **ppEnum) {
	if (!ppEnum)
		return E_POINTER;

	*ppEnum = new(std::nothrow) DSSampleGrabberMediaTypeEnumerator(mMediaType, mSubType);
	if (!*ppEnum)
		return E_OUTOFMEMORY;

	(*ppEnum)->AddRef();
	return S_OK;
}

///////////////////////////////////////////////////////////////////////////////

DSSampleGrabber::DSSampleGrabber() {
	mInputPin.SetAssociatedOutput(&mOutputPin);
	mOutputPin.SetAssociatedInput(&mInputPin);

	ReservePins(2);
	AddPin(&mInputPin, L"in");
	AddPin(&mOutputPin, L"out");
}

DSSampleGrabber::~DSSampleGrabber() {
}

void DSSampleGrabber::SetCallback(IDSSampleGrabberCallback *pCB) {
	mInputPin.SetCallback(pCB);
}

void DSSampleGrabber::SetMediaType(const GUID *type) {
	mInputPin.SetMediaType(type);
}

void DSSampleGrabber::SetMediaSubType(const GUID *subtype) {
	mInputPin.SetMediaSubType(subtype);
}

void DSSampleGrabber::SetMediaTypeFilter(IDSSampleGrabberTypeFilter *typeFilter) {
	mInputPin.SetMediaTypeFilter(typeFilter);
}

bool DSSampleGrabber::GetEstimatedStreamPosition(REFERENCE_TIME *pos) {
	return mInputPin.GetEstimatedStreamPosition(pos);
}

REFERENCE_TIME DSSampleGrabber::GetCachedStartTime() {
	return mInputPin.GetCachedStartTime();
}

REFERENCE_TIME DSSampleGrabber::GetCachedEndTime() {
	return mInputPin.GetCachedEndTime();
}

DSSampleGrabberCheckResult DSSampleGrabber::CheckForSample(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend) {
	return mInputPin.CheckForSample(tstart, tend);
}

void DSSampleGrabber::ResetOneShot(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend, bool forceAcceptLate) {
	mInputPin.ResetOneShot(tstart, tend, forceAcceptLate);
}

bool DSSampleGrabber::ResumeOneShot(const REFERENCE_TIME& tstart, const REFERENCE_TIME& tend) {
	return mInputPin.ResumeOneShot(tstart, tend);
}

void DSSampleGrabber::WaitForSample() {
	HANDLE h = mInputPin.GetSignal().getHandle();

	for(;;) {
		if (mInputPin.IsStreamEnded())
			break;

		DWORD rval = MsgWaitForMultipleObjects(1, &h, FALSE, 10000, QS_ALLEVENTS);

		if (rval == WAIT_TIMEOUT)
			continue;

		if (rval != WAIT_OBJECT_0 + 1)
			break;

		MSG msg;
		while(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);

			if (WaitForSingleObject(h, 0) == WAIT_OBJECT_0)
				return;
		}
	}
}

HRESULT STDMETHODCALLTYPE DSSampleGrabber::GetClassID(CLSID *pClassID) {
	// {F3557C75-0FE6-45fd-87D9-E1CA8D444C46}
	static const GUID CLSID_DSSampleGrabber = { 0xf3557c75, 0xfe6, 0x45fd, { 0x87, 0xd9, 0xe1, 0xca, 0x8d, 0x44, 0x4c, 0x46 } };

	*pClassID = CLSID_DSSampleGrabber;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE DSSampleGrabber::Stop() {
	mInputPin.SetRunning(false);

	return BaseFilter::Stop();
}

HRESULT STDMETHODCALLTYPE DSSampleGrabber::Pause() {
	mInputPin.SetRunning(true);

	return BaseFilter::Pause();
}

HRESULT STDMETHODCALLTYPE DSSampleGrabber::Run(REFERENCE_TIME tStart) {
	mInputPin.SetRunning(true);

	return BaseFilter::Run(tStart);
}
