// DShowInputDriver: DirectShow based input plugin for VirtualDub
// Copyright (C) 2010-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#include "stdafx.h"
#include <vd2/system/refcount.h>
#include "TransformPin.h"

TransformInputPin::TransformInputPin()
	: mbFlushing(false)
	, mpAssocOutputPin(NULL)
{
}

TransformInputPin::~TransformInputPin()
{
}

void TransformInputPin::SetAssociatedOutput(TransformOutputPin *pin) {
	mpAssocOutputPin = pin;
}

HRESULT STDMETHODCALLTYPE TransformInputPin::ReceiveConnection(IPin *pConnector, const AM_MEDIA_TYPE *pmt) {
	HRESULT hr = MemInputPin::ReceiveConnection(pConnector, pmt);

	if (FAILED(hr))
		return hr;

	if (mpAssocOutputPin)
		mpAssocOutputPin->NotifyNewInputFormat();

	return hr;
}

HRESULT STDMETHODCALLTYPE TransformInputPin::QueryAccept(const AM_MEDIA_TYPE *pmt) {
	if (mpAssocOutputPin) {
		const AM_MEDIA_TYPE *pConnectedType = mpAssocOutputPin->GetConnectedMediaType();

		if (pConnectedType)
			return DSAreMediaTypesEqual(pmt, pConnectedType);
	}

	return S_OK;
}

HRESULT STDMETHODCALLTYPE TransformInputPin::QueryInternalConnections(IPin **apPin, ULONG *nPin) {
	if (!mpAssocOutputPin)
		return MemInputPin::QueryInternalConnections(apPin, nPin);

	if (!apPin || !nPin)
		return E_POINTER;

	if (*nPin < 1)
		return S_FALSE;

	*apPin = mpAssocOutputPin;
	mpAssocOutputPin->AddRef();
	*nPin = 1;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE TransformInputPin::BeginFlush() {
	// We must NOT do this with the streaming lock held. That would cause a deadlock in case
	// we're waiting for the downstream filter to unblock, which requires a flush.
	if (mpAssocOutputPin) {
		IPin *pin = mpAssocOutputPin->GetConnectedPin();

		if (pin)
			pin->BeginFlush();
	}

	mLock.Enter();
	mbFlushing = true;
	mLock.Exit();

	return S_OK;
}

HRESULT STDMETHODCALLTYPE TransformInputPin::EndFlush() {
	mLock.Enter();
	if (mbFlushing) {
		if (mpAssocOutputPin) {
			IPin *pin = mpAssocOutputPin->GetConnectedPin();

			if (pin)
				pin->EndFlush();
		}
		mbFlushing = false;
	}
	mLock.Exit();

	return S_OK;
}

HRESULT STDMETHODCALLTYPE TransformInputPin::EndOfStream() {
	mLock.Enter();
	if (mpAssocOutputPin && !mbFlushing) {
		IPin *pin = mpAssocOutputPin->GetConnectedPin();

		if (pin)
			pin->EndOfStream();
	}
	mLock.Exit();

	return S_OK;
}

HRESULT STDMETHODCALLTYPE TransformInputPin::NewSegment(REFERENCE_TIME tStart, REFERENCE_TIME tStop, double dRate) {
	mLock.Enter();
	if (mpAssocOutputPin) {
		IPin *pin = mpAssocOutputPin->GetConnectedPin();

		if (pin)
			pin->NewSegment(tStart, tStop, dRate);
	}
	mLock.Exit();

	return S_OK;
}

///////////////////////////////////////////////////////////////////////////////

class TransformOutputMediaTypeEnumerator : public IEnumMediaTypes {
	TransformOutputMediaTypeEnumerator& operator=(const TransformOutputMediaTypeEnumerator&);
public:
	TransformOutputMediaTypeEnumerator(TransformOutputPin *pin);
	TransformOutputMediaTypeEnumerator(const TransformOutputMediaTypeEnumerator&);
	~TransformOutputMediaTypeEnumerator();

	ULONG STDMETHODCALLTYPE AddRef();
	ULONG STDMETHODCALLTYPE Release();
	HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void **ppvObject);

	HRESULT STDMETHODCALLTYPE Next(ULONG cMediaTypes, AM_MEDIA_TYPE **ppMediaTypes, ULONG *pcFetched);
	HRESULT STDMETHODCALLTYPE Skip(ULONG cMediaTypes);
	HRESULT STDMETHODCALLTYPE Reset();
	HRESULT STDMETHODCALLTYPE Clone(IEnumMediaTypes **ppEnum);

protected:
	LONG mRefCount;

	vdrefptr<TransformOutputPin> mpPin;
	int mFormatGeneration;
	int mFormatIndex;
};

TransformOutputMediaTypeEnumerator::TransformOutputMediaTypeEnumerator(TransformOutputPin *pin)
	: mRefCount(0)
	, mpPin(pin)
{
	Reset();
}

TransformOutputMediaTypeEnumerator::TransformOutputMediaTypeEnumerator(const TransformOutputMediaTypeEnumerator& src)
	: mRefCount(0)
	, mpPin(src.mpPin)
	, mFormatGeneration(src.mFormatGeneration)
	, mFormatIndex(src.mFormatIndex)
{
}

TransformOutputMediaTypeEnumerator::~TransformOutputMediaTypeEnumerator() {
}

ULONG STDMETHODCALLTYPE TransformOutputMediaTypeEnumerator::AddRef() {
	return (ULONG)InterlockedIncrement(&mRefCount);
}

ULONG STDMETHODCALLTYPE TransformOutputMediaTypeEnumerator::Release() {
	LONG rv = InterlockedDecrement(&mRefCount);

	if (!rv)
		delete this;

	return (ULONG)rv;
}

HRESULT STDMETHODCALLTYPE TransformOutputMediaTypeEnumerator::QueryInterface(REFIID riid, void **ppvObject) {
	if (riid == IID_IEnumMediaTypes) {
		*ppvObject = static_cast<IEnumMediaTypes *>(this);
		AddRef();
		return S_OK;
	}

	if (riid == IID_IUnknown) {
		*ppvObject = static_cast<IUnknown *>(this);
		AddRef();
		return S_OK;
	}

	*ppvObject = NULL;
	return E_NOINTERFACE;
}

HRESULT STDMETHODCALLTYPE TransformOutputMediaTypeEnumerator::Next(ULONG cMediaTypes, AM_MEDIA_TYPE **ppMediaTypes, ULONG *pcFetched) {
	if (!cMediaTypes)
		return S_OK;

	if (!ppMediaTypes)
		return E_POINTER;

	mpPin->mLock.Enter();
	int gen = mpPin->mFormatGeneration;
	MediaTypePtr format;
	TransformInputPin *inpin = mpPin->GetAssociatedInput();
	if (inpin)
		format = inpin->GetConnectedMediaType();
	mpPin->mLock.Exit();

	if (gen != mFormatGeneration)
		return VFW_E_ENUM_OUT_OF_SYNC;

	if (mFormatIndex == 0 && format) {
		*ppMediaTypes = format.Detach();

		mFormatIndex = 1;
		if (pcFetched)
			*pcFetched = 1;

		return S_OK;
	}

	if (pcFetched)
		*pcFetched = 0;

	return S_FALSE;
}

HRESULT STDMETHODCALLTYPE TransformOutputMediaTypeEnumerator::Skip(ULONG cMediaTypes) {
	if (!cMediaTypes)
		return S_OK;

	mpPin->mLock.Enter();
	int gen = mpPin->mFormatGeneration;
	TransformInputPin *inpin = mpPin->GetAssociatedInput();
	bool formatExists = (inpin && inpin->GetConnectedMediaType());
	mpPin->mLock.Exit();

	if (gen != mFormatGeneration)
		return VFW_E_ENUM_OUT_OF_SYNC;

	if (!formatExists)
		return S_FALSE;

	mFormatIndex = 1;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE TransformOutputMediaTypeEnumerator::Reset() {
	mpPin->mLock.Enter();
	mFormatGeneration = mpPin->mFormatGeneration;
	mFormatIndex = 0;
	mpPin->mLock.Exit();
	return S_OK;
}

HRESULT STDMETHODCALLTYPE TransformOutputMediaTypeEnumerator::Clone(IEnumMediaTypes **ppEnum) {
	if (!ppEnum)
		return E_POINTER;

	*ppEnum = new(std::nothrow) TransformOutputMediaTypeEnumerator(*this);
	if (!*ppEnum)
		return E_OUTOFMEMORY;

	(*ppEnum)->AddRef();
	return S_OK;
}

///////////////////////////////////////////////////////////////////////////////

TransformOutputPin::TransformOutputPin() {
}

TransformOutputPin::~TransformOutputPin() {
}

void TransformOutputPin::SetAssociatedInput(TransformInputPin *pin) {
	mpAssocInputPin = pin;
}

void TransformOutputPin::NotifyNewInputFormat() {
	++mFormatGeneration;
}

void TransformOutputPin::Output(IMediaSample *sample) {
	mLock.Enter();
	if (mpTransportPin)
		mpTransportPin->Receive(sample);
	mLock.Exit();
}

ULONG STDMETHODCALLTYPE TransformOutputPin::AddRef() {
	return MemOutputPin::AddRef();
}

ULONG STDMETHODCALLTYPE TransformOutputPin::Release() {
	return MemOutputPin::Release();
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::QueryInterface(REFIID riid, void **ppvObject) {
	if (riid == IID_IMediaSeeking) {
		*ppvObject = static_cast<IMediaSeeking *>(this);
		AddRef();
		return S_OK;
	}

	return MemOutputPin::QueryInterface(riid, ppvObject);
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::QueryInternalConnections(IPin **apPin, ULONG *nPin) {
	if (!apPin || !nPin)
		return E_POINTER;

	if (*nPin < 1)
		return S_FALSE;

	*apPin = mpAssocInputPin;
	mpAssocInputPin->AddRef();
	*nPin = 1;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::EnumMediaTypes(IEnumMediaTypes **ppEnum) {
	if (!ppEnum)
		return E_POINTER;

	*ppEnum = new(std::nothrow) TransformOutputMediaTypeEnumerator(this);
	if (!*ppEnum)
		return E_OUTOFMEMORY;
	(*ppEnum)->AddRef();
	return S_OK;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::GetCapabilities(DWORD *pCapabilities) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->GetCapabilities(pCapabilities);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::CheckCapabilities(DWORD *pCapabilities) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->CheckCapabilities(pCapabilities);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::IsFormatSupported(const GUID *pFormat) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->IsFormatSupported(pFormat);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::QueryPreferredFormat(GUID *pFormat) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->QueryPreferredFormat(pFormat);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::GetTimeFormat(GUID *pFormat) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->GetTimeFormat(pFormat);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::IsUsingTimeFormat(const GUID *pFormat) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->IsUsingTimeFormat(pFormat);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::SetTimeFormat(const GUID *pFormat) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->SetTimeFormat(pFormat);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::GetDuration(LONGLONG *pDuration) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->GetDuration(pDuration);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::GetStopPosition(LONGLONG *pStop) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->GetStopPosition(pStop);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::GetCurrentPosition(LONGLONG *pCurrent) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->GetCurrentPosition(pCurrent);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::ConvertTimeFormat(LONGLONG *pTarget, const GUID *pTargetFormat, LONGLONG Source, const GUID *pSourceFormat) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->ConvertTimeFormat(pTarget, pTargetFormat, Source, pSourceFormat);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::SetPositions(LONGLONG *pCurrent, DWORD dwCurrentFlags, LONGLONG *pStop, DWORD dwStopFlags) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->SetPositions(pCurrent, dwCurrentFlags, pStop, dwStopFlags);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::GetPositions(LONGLONG *pCurrent, LONGLONG *pStop) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->GetPositions(pCurrent, pStop);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::GetAvailable(LONGLONG *pEarliest, LONGLONG *pLatest) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->GetAvailable(pEarliest, pLatest);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::SetRate(double dRate) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->SetRate(dRate);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::GetRate(double *pdRate) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->GetRate(pdRate);
	pMS->Release();
	return hr;
}

HRESULT STDMETHODCALLTYPE TransformOutputPin::GetPreroll(LONGLONG *pllPreroll) {
	IMediaSeeking *pMS;
	if (!GetUpstreamMediaSeeking(&pMS))
		return E_NOTIMPL;

	HRESULT hr = pMS->GetPreroll(pllPreroll);
	pMS->Release();
	return hr;
}

bool TransformOutputPin::GetUpstreamMediaSeeking(IMediaSeeking **ppMS) {
	IPin *pin = mpAssocInputPin->GetConnectedPin();
	if (!pin)
		return false;

	HRESULT hr = pin->QueryInterface(IID_IMediaSeeking, (void **)ppMS);
	return SUCCEEDED(hr);
}
