// DShowInputDriver: DirectShow based input plugin for VirtualDub
// Copyright (C) 2010-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#include "stdafx.h"
#include <vd2/system/strutil.h>
#include <vd2/system/vdtypes.h>
#include <vd2/system/vectors.h>
#include <vd2/system/w32assist.h>
#include "resource.h"
#include "Config.h"
#include <windows.h>
#include <commctrl.h>

///////////////////////////////////////////////////////////////////////////

class UIDialogBase {
public:
	UIDialogBase();

	void SetPosition(const vdpoint32& pt);
	vdsize32 GetSize() const;

	void Show(bool visible);

protected:
	static INT_PTR CALLBACK StaticDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam);
	virtual INT_PTR DlgProc(UINT msg, WPARAM wParam, LPARAM lParam) = 0;

	HWND mhdlg;

};

UIDialogBase::UIDialogBase()
	: mhdlg(NULL)
{
}

void UIDialogBase::SetPosition(const vdpoint32& pt) {
	if (mhdlg)
		SetWindowPos(mhdlg, NULL, pt.x, pt.y, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
}

vdsize32 UIDialogBase::GetSize() const {
	if (mhdlg) {
		RECT r;
		if (GetWindowRect(mhdlg, &r))
			return vdsize32(r.right - r.left, r.bottom - r.top);
	}

	return vdsize32(0, 0);
}

void UIDialogBase::Show(bool vis) {
	if (mhdlg)
		ShowWindow(mhdlg, vis ? SW_SHOWNOACTIVATE : SW_HIDE);
}

INT_PTR CALLBACK UIDialogBase::StaticDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	UIDialogBase *p;

	if (msg == WM_INITDIALOG) {
		SetWindowLongPtr(hdlg, DWLP_USER, (LONG_PTR)lParam);

		p = (UIDialogBase *)lParam;
		p->mhdlg = hdlg;
	} else {
		p = (UIDialogBase *)GetWindowLongPtr(hdlg, DWLP_USER);
	}

	return p ? p->DlgProc(msg, wParam, lParam) : FALSE;
}

///////////////////////////////////////////////////////////////////////////

class UIDialogConfigAudio : public UIDialogBase {
public:
	HWND Create(HWND hwndParent);
	void Load(VDRegistryKey& key);
	void Commit(VDRegistryKey& key);

protected:
	INT_PTR DlgProc(UINT msg, WPARAM wParam, LPARAM lParam);
	void OnInitDialog();
};

HWND UIDialogConfigAudio::Create(HWND hwndParent) {
	HINSTANCE hInst = VDGetLocalModuleHandleW32();

	return CreateDialogParam(hInst, MAKEINTRESOURCE(IDD_CONFIGURE_AUDIO), hwndParent, StaticDlgProc, (LPARAM)this);
}

void UIDialogConfigAudio::Load(VDRegistryKey& key) {
	CheckDlgButton(mhdlg, IDC_FORCE_STEREO, key.getBool("Audio: Force stereo", true));
}

void UIDialogConfigAudio::Commit(VDRegistryKey& key) {
	key.setBool("Audio: Force stereo", IsDlgButtonChecked(mhdlg, IDC_FORCE_STEREO) != 0);
}

INT_PTR UIDialogConfigAudio::DlgProc(UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_INITDIALOG:
			OnInitDialog();
			break;
	}

	return FALSE;
}

void UIDialogConfigAudio::OnInitDialog() {
}

///////////////////////////////////////////////////////////////////////////

class UIDialogConfigVideo : public UIDialogBase {
public:
	HWND Create(HWND hwndParent);
	void Load(VDRegistryKey& key);
	void Commit(VDRegistryKey& key);

protected:
	INT_PTR DlgProc(UINT msg, WPARAM wParam, LPARAM lParam);
	void OnInitDialog();
};

HWND UIDialogConfigVideo::Create(HWND hwndParent) {
	HINSTANCE hInst = VDGetLocalModuleHandleW32();

	return CreateDialogParam(hInst, MAKEINTRESOURCE(IDD_CONFIGURE_VIDEO), hwndParent, StaticDlgProc, (LPARAM)this);
}

void UIDialogConfigVideo::Load(VDRegistryKey& key) {
	CheckDlgButton(mhdlg, IDC_ENDPOSWORKAROUND, key.getBool("Video: Use end position workaround", false));
}

void UIDialogConfigVideo::Commit(VDRegistryKey& key) {
	key.setBool("Video: Use end position workaround", IsDlgButtonChecked(mhdlg, IDC_ENDPOSWORKAROUND) != 0);
}

INT_PTR UIDialogConfigVideo::DlgProc(UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_INITDIALOG:
			OnInitDialog();
			break;
	}

	return FALSE;
}

void UIDialogConfigVideo::OnInitDialog() {
}

///////////////////////////////////////////////////////////////////////////

class UIDialogConfigFileTypes : public UIDialogBase {
public:
	HWND Create(HWND hwndParent);
	void Load(VDRegistryKey& key);
	void Commit(VDRegistryKey& key);

protected:
	INT_PTR DlgProc(UINT msg, WPARAM wParam, LPARAM lParam);
	bool OnInitDialog();
	bool OnCommand(uint32 id);

	const char *GetString(size_t idx);

	HWND mhwndList;
	HWND mhwndEdit;
	VDStringA mText;
};

HWND UIDialogConfigFileTypes::Create(HWND hwndParent) {
	HINSTANCE hInst = VDGetLocalModuleHandleW32();

	return CreateDialogParam(hInst, MAKEINTRESOURCE(IDD_CONFIGURE_FILETYPES), hwndParent, StaticDlgProc, (LPARAM)this);
}

void UIDialogConfigFileTypes::Load(VDRegistryKey& key) {
	bool matchReg = true;

	mhwndList = GetDlgItem(mhdlg, IDC_LIST);
	mhwndEdit = GetDlgItem(mhdlg, IDC_EDIT);

	if (mhwndList) {
		VDStringA s;
		if (key.getString("Extensions", s)) {
			VDStringRefA token;
			VDStringRefA sr(s);

			while(!sr.empty()) {
				if (!sr.split(';', token)) {
					token = sr;
					sr.clear();
				}

				if (!token.empty())
					SendMessageA(mhwndList, LB_ADDSTRING, 0, (LPARAM)VDStringA(token).c_str());
			}
		}

		matchReg = key.getBool("Match registered", matchReg);
	}

	CheckDlgButton(mhdlg, IDC_MATCHREGISTERED, matchReg ? BST_CHECKED : BST_UNCHECKED);
}

void UIDialogConfigFileTypes::Commit(VDRegistryKey& key) {
	if (mhwndList) {
		VDStringA s;

		int n = SendMessage(mhwndList, LB_GETCOUNT, 0, 0);
		if (n > 0) {
			for(int i=0; i<n; ++i) {
				const char *listExt = GetString(i);

				if (*listExt) {
					if (!s.empty())
						s += ';';

					s.append(listExt);
				}
			}
		}

		key.setString("Extensions", s.c_str());
	}

	key.setBool("Match registered", IsDlgButtonChecked(mhdlg, IDC_MATCHREGISTERED) != 0);
}

INT_PTR UIDialogConfigFileTypes::DlgProc(UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_INITDIALOG:
			return OnInitDialog();

		case WM_COMMAND:
			return OnCommand(LOWORD(wParam));
	}

	return FALSE;
}

bool UIDialogConfigFileTypes::OnInitDialog() {
	SetFocus(GetDlgItem(mhdlg, IDC_MATCHREGISTERED));
	return true;
}

bool UIDialogConfigFileTypes::OnCommand(uint32 id) {
	switch(id) {
		case IDC_ADD:
			if (mhwndEdit && mhwndList) {
				VDStringA ext(VDGetWindowTextAW32(mhwndEdit));

				if (ext.empty())
					return true;

				if (ext[0] == '.')
					ext.erase(0, 1);

				if (ext.empty())
					return true;

				int n = SendMessage(mhwndList, LB_GETCOUNT, 0, 0);
				if (n > 0) {
					for(int i=0; i<n; ++i) {
						const char *listExt = GetString(i);

						if (!vdstricmp(listExt, ext.c_str()))
							return true;
					}
				}

				SendMessageA(mhwndList, LB_ADDSTRING, 0, (LPARAM)ext.c_str());
			}
			return true;

		case IDC_DELETE:
			if (mhwndList) {
				int idx = SendMessage(mhwndList, LB_GETCURSEL, 0, 0);

				if (idx >= 0)
					SendMessage(mhwndList, LB_DELETESTRING, idx, 0);
			}
			return true;
	}

	return false;
}

const char *UIDialogConfigFileTypes::GetString(size_t idx) {
	mText.clear();

	if (mhwndList) {
		int len = SendMessageA(mhwndList, LB_GETTEXTLEN, idx, 0);

		if (len >= 0) {
			mText.resize(len + 1);

			len = SendMessageA(mhwndList, LB_GETTEXT, idx, (LPARAM)(char *)mText.data());

			if (len < 0)
				mText.clear();
			else
				mText.resize(len);
		}
	}

	return mText.c_str();
}
///////////////////////////////////////////////////////////////////////////

class UIDialogConfig : public UIDialogBase {
public:
	void ShowDialog(HWND hwndParent);

protected:
	INT_PTR DlgProc(UINT msg, WPARAM wParam, LPARAM lParam);
	void OnInitDialog();
	bool OnCommand(uint32 id);
	void ShowTab(int idx);
	void Load();
	void Commit();

	HWND mhwndTabCtrl;
	UIDialogConfigAudio mPageAudio;
	UIDialogConfigVideo mPageVideo;
	UIDialogConfigFileTypes mPageFileTypes;
};

void UIDialogConfig::ShowDialog(HWND hwndParent) {
	HINSTANCE hInst = VDGetLocalModuleHandleW32();

	DialogBoxParam(hInst, MAKEINTRESOURCE(IDD_CONFIGURE), hwndParent, StaticDlgProc, (LPARAM)this);
}

INT_PTR UIDialogConfig::DlgProc(UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
		case WM_INITDIALOG:
			OnInitDialog();
			break;

		case WM_COMMAND:
			if (OnCommand(LOWORD(wParam)))
				return TRUE;
			break;

		case WM_NOTIFY:
			{
				const NMHDR& hdr = *(const NMHDR *)lParam;

				if (hdr.hwndFrom == mhwndTabCtrl && hdr.code == TCN_SELCHANGE) {
					ShowTab(TabCtrl_GetCurSel(mhwndTabCtrl));
				}
			}
			break;
	}
	return FALSE;
}

void UIDialogConfig::OnInitDialog() {
	mPageAudio.Create(mhdlg);
	mPageVideo.Create(mhdlg);
	mPageFileTypes.Create(mhdlg);

	vdsize32 pageSize(0, 0);
	pageSize.include(mPageAudio.GetSize());
	pageSize.include(mPageVideo.GetSize());
	pageSize.include(mPageFileTypes.GetSize());

	mhwndTabCtrl = GetDlgItem(mhdlg, IDC_TABS);
	if (mhwndTabCtrl) {
		TCITEM tci = {};
		tci.mask = TCIF_TEXT;
		tci.pszText = _T("File Types");
		TabCtrl_InsertItem(mhwndTabCtrl, 0, &tci);

		tci.pszText = _T("Audio");
		TabCtrl_InsertItem(mhwndTabCtrl, 1, &tci);

		tci.pszText = _T("Video");
		TabCtrl_InsertItem(mhwndTabCtrl, 2, &tci);

		RECT rOld = {0};
		GetWindowRect(mhwndTabCtrl, &rOld);
		MapWindowPoints(NULL, mhdlg, (LPPOINT)&rOld, 2);

		RECT rNew = {0, 0, pageSize.w, pageSize.h};
		TabCtrl_AdjustRect(mhwndTabCtrl, TRUE, &rNew);

		const int nx = rNew.right - rNew.left;
		const int ny = rNew.bottom - rNew.top;
		const int dx = nx - (rOld.right - rOld.left);
		const int dy = ny - (rOld.bottom - rOld.top);

		vdpoint32 pagePos(rOld.left - rNew.left, rOld.top - rNew.top);
		mPageAudio.SetPosition(pagePos);
		mPageVideo.SetPosition(pagePos);
		mPageFileTypes.SetPosition(pagePos);

		HDWP hdwp = BeginDeferWindowPos(3);
		if (hdwp) {
			RECT rt = {0};
			HDWP hdwp2;

			hdwp2 = DeferWindowPos(hdwp, mhwndTabCtrl, NULL, 0, 0, nx, ny, SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);
			if (hdwp2)
				hdwp = hdwp2;

			HWND hwndItem = GetDlgItem(mhdlg, IDOK);
			if (hwndItem) {
				GetWindowRect(hwndItem, &rt);
				MapWindowPoints(NULL, mhdlg, (LPPOINT)&rt, 2);
				hdwp2 = DeferWindowPos(hdwp, hwndItem, NULL, rt.left + dx, rt.top + dy, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
				if (hdwp2)
					hdwp = hdwp2;
			}

			hwndItem = GetDlgItem(mhdlg, IDCANCEL);
			if (hwndItem) {
				GetWindowRect(hwndItem, &rt);
				MapWindowPoints(NULL, mhdlg, (LPPOINT)&rt, 2);
				hdwp2 = DeferWindowPos(hdwp, hwndItem, NULL, rt.left + dx, rt.top + dy, 0, 0, SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
				if (hdwp2)
					hdwp = hdwp2;
			}

			EndDeferWindowPos(hdwp);
		}

		RECT rDlg;
		if (GetWindowRect(mhdlg, &rDlg))
			SetWindowPos(mhdlg, NULL, 0, 0, rDlg.right - rDlg.left + dx, rDlg.bottom - rDlg.top + dy, SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE);
	}

	ShowTab(0);
	Load();
}

bool UIDialogConfig::OnCommand(uint32 id) {
	switch(id) {
		case IDOK:
			Commit();
			EndDialog(mhdlg, TRUE);
			return true;

		case IDCANCEL:
			EndDialog(mhdlg, FALSE);
			return true;
	}

	return false;
}

void UIDialogConfig::ShowTab(int idx) {
	if (mhwndTabCtrl)
		TabCtrl_SetCurSel(mhwndTabCtrl, idx);

	mPageFileTypes.Show(idx == 0);
	mPageAudio.Show(idx == 1);
	mPageVideo.Show(idx == 2);
}

void UIDialogConfig::Load() {
	DSConfigRegistryKey key(false);

	mPageAudio.Load(key);
	mPageVideo.Load(key);
	mPageFileTypes.Load(key);
}

void UIDialogConfig::Commit() {
	DSConfigRegistryKey key(true);

	mPageAudio.Commit(key);
	mPageVideo.Commit(key);
	mPageFileTypes.Commit(key);
}

///////////////////////////////////////////////////////////////////////////

void ShowUIDialogConfig(VDGUIHandle hwnd) {
	UIDialogConfig dlg;

	dlg.ShowDialog((HWND)hwnd);
}
