// DShowInputDriver: DirectShow based input plugin for VirtualDub
// Copyright (C) 2010-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#include "stdafx.h"
#include <vd2/system/refcount.h>
#include <vd2/system/VDString.h>
#include "Utils.h"

///////////////////////////////////////////////////////////////////////////////

MediaTypePtr::MediaTypePtr()
	: mpMediaType(NULL)
{
}

MediaTypePtr::MediaTypePtr(const AM_MEDIA_TYPE *src)
	: mpMediaType(NULL)
{
	operator=(src);
}

MediaTypePtr::MediaTypePtr(const MediaTypePtr& src)
	: mpMediaType(NULL)
{
	operator=(src);
}

MediaTypePtr::~MediaTypePtr() {
	operator=(NULL);
}

MediaTypePtr& MediaTypePtr::operator=(const AM_MEDIA_TYPE *p) {
	if (p == mpMediaType)
		return *this;

	if (p) {
		if (!mpMediaType) {
			mpMediaType = (AM_MEDIA_TYPE *)CoTaskMemAlloc(sizeof(AM_MEDIA_TYPE));
			if (!mpMediaType)
				return *this;

			memset(mpMediaType, 0, sizeof(AM_MEDIA_TYPE));
		}

		if (mpMediaType->pbFormat) {
			CoTaskMemFree(mpMediaType->pbFormat);
			mpMediaType->pbFormat = NULL;
		}
		mpMediaType->cbFormat = 0;

		if (p->pbFormat) {
			mpMediaType->pbFormat = (BYTE *)CoTaskMemAlloc(p->cbFormat);
			if (!mpMediaType->pbFormat) {
				CoTaskMemFree(mpMediaType);
				mpMediaType = NULL;
				return *this;
			}

			memcpy(mpMediaType->pbFormat, p->pbFormat, p->cbFormat);
			mpMediaType->cbFormat = p->cbFormat;
		}

		mpMediaType->majortype				= p->majortype;
		mpMediaType->subtype				= p->subtype;
		mpMediaType->bFixedSizeSamples		= p->bFixedSizeSamples;
		mpMediaType->bTemporalCompression	= p->bTemporalCompression;
		mpMediaType->lSampleSize			= p->lSampleSize;
		mpMediaType->formattype				= p->formattype;

		if (p->pUnk)
			p->pUnk->AddRef();
		if (mpMediaType->pUnk)
			mpMediaType->pUnk->Release();
		mpMediaType->pUnk = p->pUnk;
	} else {
		if (mpMediaType) {
			if (mpMediaType->pbFormat)
				CoTaskMemFree(mpMediaType->pbFormat);

			if (mpMediaType->pUnk)
				mpMediaType->pUnk->Release();

			CoTaskMemFree(mpMediaType);
			mpMediaType = NULL;
		}
	}

	return *this;
}

MediaTypePtr& MediaTypePtr::operator=(const MediaTypePtr& p) {
	return operator=(static_cast<const AM_MEDIA_TYPE *>(p));
}

void MediaTypePtr::Attach(AM_MEDIA_TYPE *type) {
	if (type == mpMediaType)
		return;

	operator=(NULL);

	mpMediaType = type;
}

AM_MEDIA_TYPE *MediaTypePtr::Detach() {
	AM_MEDIA_TYPE *p = mpMediaType;
	mpMediaType = NULL;

	return p;
}

///////////////////////////////////////////////////////////////////////////////

Lock::Lock() {
	InitializeCriticalSection(&mCsect);
}

Lock::Lock(const Lock&) {
	InitializeCriticalSection(&mCsect);
}

Lock::~Lock() {
	DeleteCriticalSection(&mCsect);
}

void Lock::Enter() {
	EnterCriticalSection(&mCsect);
}

void Lock::Exit() {
	LeaveCriticalSection(&mCsect);
}

///////////////////////////////////////////////////////////////////////////////

bool DSAreMediaTypesEqual(const AM_MEDIA_TYPE *p, const AM_MEDIA_TYPE *q) {
	if (p->majortype != q->majortype)
		return false;

	if (p->subtype != q->subtype)
		return false;

	if (p->bFixedSizeSamples != q->bFixedSizeSamples)
		return false;

	if (p->bTemporalCompression != q->bTemporalCompression)
		return false;

	if (p->lSampleSize != q->lSampleSize)
		return false;

	if (p->formattype != q->formattype)
		return false;

	// ignore pUnk

	if (p->cbFormat != q->cbFormat)
		return false;

	if (q->pbFormat) {
		if (!p->pbFormat || memcmp(q->pbFormat, p->pbFormat, p->cbFormat))
			return false;
	}

	return true;
}

///////////////////////////////////////////////////////////////////////////////

bool DSAddToRot(IUnknown *pUnkGraph, DWORD *pdwRegister) {
	IRunningObjectTable *pROT;
	HRESULT hr = GetRunningObjectTable(0, &pROT);
	if (FAILED(hr))
		return false;

	WCHAR wsz[256];
	wsprintfW(wsz, L"FilterGraph %08x pid %08x", (DWORD_PTR)pUnkGraph, GetCurrentProcessId());
	IMoniker *pMoniker;
	hr = CreateItemMoniker(L"!", wsz, &pMoniker);
	if (SUCCEEDED(hr)) {
		hr = pROT->Register(ROTFLAGS_REGISTRATIONKEEPSALIVE, pUnkGraph, pMoniker, pdwRegister);
		pMoniker->Release();
	}
	pROT->Release();
	return SUCCEEDED(hr);
}

void DSRemoveFromRot(DWORD pdwRegister) {
    IRunningObjectTable *pROT;
	HRESULT hr = GetRunningObjectTable(0, &pROT);
	if (SUCCEEDED(hr)) {
		pROT->Revoke(pdwRegister);
		pROT->Release();
	}
}

void DSListFilters(IFilterGraph *pGraph, VDStringA& s) {
	vdrefptr<IEnumFilters> pEnumFilters;

	HRESULT hr = pGraph->EnumFilters(~pEnumFilters);
	if (FAILED(hr))
		return;

	vdrefptr<IBaseFilter> filter;
	ULONG fetched;
	while(S_OK == pEnumFilters->Next(1, ~filter, &fetched)) {
		FILTER_INFO fi = {0};
		hr = filter->QueryFilterInfo(&fi);
		if (SUCCEEDED(hr)) {
			if (fi.pGraph)
				fi.pGraph->Release();

			s += "  Filter\t";
			s += VDTextWToA(fi.achName);

			LPWSTR name = NULL;
			hr = filter->QueryVendorInfo(&name);
			if (SUCCEEDED(hr) && name) {
				s += '(';
				s += VDTextWToA(name);
				s += ')';

				CoTaskMemFree(name);
			}

			s += '\n';
		}
	}
}
