// DShowInputDriver: DirectShow based input plugin for VirtualDub
// Copyright (C) 2010-2011 Avery Lee
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#include "stdafx.h"
#include <dxva.h>
#include <vd2/system/math.h>
#include <vd2/system/fraction.h>
#include <vd2/VDXFrame/Unknown.h>
#include "Config.h"
#include "Engine.h"
#include "SampleGrabber.h"
#include "VideoSourceDShow.h"

DEFINE_GUID(DXVA_ModeH264_A,  0x1b81be64, 0xa0c7,0x11d3,0xb9,0x84,0x00,0xc0,0x4f,0x2e,0x73,0xc5);
DEFINE_GUID(DXVA_ModeH264_B,  0x1b81be65, 0xa0c7,0x11d3,0xb9,0x84,0x00,0xc0,0x4f,0x2e,0x73,0xc5);
DEFINE_GUID(DXVA_ModeH264_C,  0x1b81be66, 0xa0c7,0x11d3,0xb9,0x84,0x00,0xc0,0x4f,0x2e,0x73,0xc5);
DEFINE_GUID(DXVA_ModeH264_D,  0x1b81be67, 0xa0c7,0x11d3,0xb9,0x84,0x00,0xc0,0x4f,0x2e,0x73,0xc5);
DEFINE_GUID(DXVA_ModeH264_E,  0x1b81be68, 0xa0c7,0x11d3,0xb9,0x84,0x00,0xc0,0x4f,0x2e,0x73,0xc5);
DEFINE_GUID(DXVA_ModeH264_F,  0x1b81be69, 0xa0c7,0x11d3,0xb9,0x84,0x00,0xc0,0x4f,0x2e,0x73,0xc5);

namespace {
	// We start at 0.5 seconds and limit reseeking to 16 seconds.
	const sint64 kSeekFudgeOffsetMin = 5000000;
	const sint64 kSeekFudgeOffsetMax = 160000000;

	void AppendVideoFormatName(VDStringA& s, const AM_MEDIA_TYPE *amt) {
		if (amt->subtype == MEDIASUBTYPE_RGB555) {
			s += "RGB555";
			return;
		} else if (amt->subtype == MEDIASUBTYPE_RGB565) {
			s += "RGB565";
			return;
		} else if (amt->subtype == MEDIASUBTYPE_RGB24) {
			s += "RGB24";
			return;
		} else if (amt->subtype == MEDIASUBTYPE_RGB32) {
			s += "RGB32";
			return;
		} else if (amt->subtype == DXVA_ModeMPEG2_A) {
			s += "DXVA_ModeMPEG2_A";
			return;
		} else if (amt->subtype == DXVA_ModeMPEG2_B) {
			s += "DXVA_ModeMPEG2_B";
			return;
		} else if (amt->subtype == DXVA_ModeMPEG2_C) {
			s += "DXVA_ModeMPEG2_C";
			return;
		} else if (amt->subtype == DXVA_ModeMPEG2_D) {
			s += "DXVA_ModeMPEG2_D";
			return;
		} else if (amt->subtype == DXVA_ModeH264_A) {
			s += "DXVA_ModeH264_A";
			return;
		} else if (amt->subtype == DXVA_ModeH264_B) {
			s += "DXVA_ModeH264_B";
			return;
		} else if (amt->subtype == DXVA_ModeH264_C) {
			s += "DXVA_ModeH264_C";
			return;
		} else if (amt->subtype == DXVA_ModeH264_D) {
			s += "DXVA_ModeH264_D";
			return;
		} else if (
			amt->subtype.Data2 == 0x0000 &&
			amt->subtype.Data3 == 0x0010 &&
			amt->subtype.Data4[0] == 0x80 &&
			amt->subtype.Data4[1] == 0x00 &&
			amt->subtype.Data4[2] == 0x00 &&
			amt->subtype.Data4[3] == 0xaa &&
			amt->subtype.Data4[4] == 0x00 &&
			amt->subtype.Data4[5] == 0x38 &&
			amt->subtype.Data4[6] == 0x9b &&
			amt->subtype.Data4[7] == 0x71)
		{
			const uint8 c1 = (uint8)amt->subtype.Data1;
			const uint8 c2 = (uint8)(amt->subtype.Data1 >> 8);
			const uint8 c3 = (uint8)(amt->subtype.Data1 >> 16);
			const uint8 c4 = (uint8)(amt->subtype.Data1 >> 24);

			if ((uint8)(c1 - 0x20) < 0x5f &&
				(uint8)(c2 - 0x20) < 0x5f &&
				(uint8)(c3 - 0x20) < 0x5f &&
				(uint8)(c4 - 0x20) < 0x5f)
			{
				s += (char)c1;
				s += (char)c2;
				s += (char)c3;
				s += (char)c4;
				return;
			}
		}

		s.append_sprintf("{%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X}"
			, amt->subtype.Data1
			, amt->subtype.Data2
			, amt->subtype.Data3
			, amt->subtype.Data4[0]
			, amt->subtype.Data4[1]
			, amt->subtype.Data4[2]
			, amt->subtype.Data4[3]
			, amt->subtype.Data4[4]
			, amt->subtype.Data4[5]
			, amt->subtype.Data4[6]
			, amt->subtype.Data4[7]);
	}
}

DSVideoDecoder::DSVideoDecoder(const VDXInputDriverContext *context, IDSEngine *engine, const wchar_t *fn)
	: mpEngine(engine)
	, mpContext(context)
	, mFormatSupportMask(0)
	, mbFrameBufferValid(false)
	, mSeekFudgeOffset(0)
	, mGraphRegistration(0)
{
	memset(&mStreamInfo, 0, sizeof mStreamInfo);
	memset(&mVideoInfo, 0, sizeof mVideoInfo);
	memset(&mpxFrameBuffer, 0, sizeof mpxFrameBuffer);

	DSConfigRegistryKey key;
	mbUseEndPosHack = key.getBool("Video: Use end position workaround", false);

	try {
		DSEngineCommand cmd(&mCommandSignal);
		cmd.SetMethodCallback<DSVideoDecoder, &DSVideoDecoder::InitLocal>(this);
		cmd.SetExtraData((void *)fn);
		mpEngine->SendCommand(&cmd);
	} catch(const MyError&) {
		ShutdownLocal(NULL);
		throw;
	}
}

void DSVideoDecoder::InitLocal(void *pvFn) {
	const wchar_t *fn = (const wchar_t *)pvFn;

	mInfo = "Video stream:\n";

	HRESULT hr = CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER, IID_IGraphBuilder, (void **)~mpGraph);
	if (FAILED(hr))
		throw MyError("Unable to create DirectShow filter graph: hr=%08x", hr);

	DSAddToRot(mpGraph, &mGraphRegistration);

	hr = mpGraph->QueryInterface(IID_IMediaControl, (void **)~mpGraphMediaControl);
	if (FAILED(hr))
		throw MyError("Unable to obtain DirectShow IMediaControl interface: hr=%08x", hr);

	hr = mpGraph->QueryInterface(IID_IMediaEvent, (void **)~mpGraphMediaEvent);
	if (FAILED(hr))
		throw MyError("Unable to obtain DirectShow IMediaEvent interface: hr=%08x", hr);

	hr = mpGraph->QueryInterface(IID_IMediaFilter, (void **)~mpGraphMediaFilter);
	if (FAILED(hr))
		throw MyError("Unable to obtain DirectShow IMediaFilter interface: hr=%08x", hr);

	mpSampleGrabber = new_nothrow DSSampleGrabber;
	if (!mpSampleGrabber)
		throw MyMemoryError();

	mpSampleGrabber->SetCallback(this);
	mpSampleGrabber->SetMediaType(&MEDIATYPE_Video);

	hr = mpGraph->AddFilter(mpSampleGrabber, L"video grabber");
	if (FAILED(hr))
		throw MyError("Unable to add sample grabber to graph: hr=%08x", hr);

	vdrefptr<ICaptureGraphBuilder2> pCB2;
	hr = CoCreateInstance(CLSID_CaptureGraphBuilder2, NULL, CLSCTX_INPROC_SERVER, IID_ICaptureGraphBuilder2, (void **)~pCB2);
	if (FAILED(hr))
		throw MyError("Unable to create capture graph builder: hr=%08x", hr);

	hr = pCB2->SetFiltergraph(mpGraph);
	if (FAILED(hr))
		throw MyError("Unable to set filter graph on capture graph builder: hr=%08x", hr);

	vdrefptr<IBaseFilter> pSourceFilter;
	hr = mpGraph->AddSourceFilter(fn, L"source", ~pSourceFilter);
	if (FAILED(hr))
		throw MyError("Unable to create source filter: hr=%08x", hr);

	hr = CoCreateInstance(CLSID_NullRenderer, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (void **)~mpRenderer);
	if (FAILED(hr))
		throw MyError("Unable to create null renderer: hr=%08x", hr);

	hr = mpGraph->AddFilter(mpRenderer, L"null renderer");
	if (FAILED(hr))
		throw MyError("Unable to add null renderer to graph: hr=%08x", hr);

	{
		vdrefptr<IPin> srcpin;
		hr = pCB2->FindPin(pSourceFilter, PINDIR_OUTPUT, NULL, &MEDIATYPE_Video, TRUE, 0, ~srcpin);
		if (FAILED(hr)) {
			hr = pCB2->RenderStream(NULL, NULL, pSourceFilter, mpSampleGrabber, mpRenderer);
			if (FAILED(hr))
				throw MyError("Unable to render video stream: hr=%08x", hr);
		} else {
			hr = mpGraph->Connect(srcpin, mpSampleGrabber->GetInputPin());
			if (FAILED(hr))
				throw MyError("Unable to render video stream: hr=%08x", hr);

			vdrefptr<IPin> rdrpin;
			hr = pCB2->FindPin(mpRenderer, PINDIR_INPUT, NULL, NULL, TRUE, 0, ~rdrpin);
			if (FAILED(hr))
				throw MyError("Unable to render video stream: hr=%08x", hr);

			hr = mpGraph->Connect(mpSampleGrabber->GetOutputPin(), rdrpin);
			if (FAILED(hr))
				throw MyError("Unable to render video stream: hr=%08x", hr);
		}
	}

	hr = mpSampleGrabber->GetInputPin()->ConnectedTo(~mpSourcePin);
	if (FAILED(hr))
		throw MyError("Unable to find source/decoder output pin: hr=%08x", hr);

	hr = mpSampleGrabber->GetOutputPin()->ConnectedTo(~mpRendererPin);
	if (FAILED(hr))
		throw MyError("Unable to find renderer input pin: hr=%08x", hr);

	// kill the graph clock
	hr = mpGraphMediaFilter->SetSyncSource(NULL);
	if (FAILED(hr))
		throw MyError("Unable to disable DirectShow filter graph clock: hr=%08x", hr);

	// get duration
	hr = mpGraph->QueryInterface(IID_IMediaSeeking, (void **)~mpGraphMediaSeeking);
	if (FAILED(hr))
		throw MyError("Unable to obtain DirectShow IMediaSeeking interface: hr=%08x", hr);

	hr = mpGraphMediaSeeking->IsFormatSupported(&TIME_FORMAT_MEDIA_TIME);
	if (hr != S_OK)
		throw MyError("Video graph doesn't support media time format: hr=%08x", hr);

	hr = mpGraphMediaSeeking->SetTimeFormat(&TIME_FORMAT_MEDIA_TIME);
	if (FAILED(hr))
		throw MyError("Unable to change video seek time format: hr=%08x", hr);

	// must get duration before we disconnect the source pin
	LONGLONG dur;
	hr = mpGraphMediaSeeking->GetDuration(&dur);
	if (FAILED(hr))
		throw MyError("Unable to get stream duration: hr=%08x", hr);

	mInfo.append_sprintf("  Duration\t%.7f seconds\n", (double)dur / 10000000.0);

	mInfo.append("  Supported seek modes\t");
	if (S_OK == mpGraphMediaSeeking->IsFormatSupported(&TIME_FORMAT_BYTE)) mInfo += "byte ";
	if (S_OK == mpGraphMediaSeeking->IsFormatSupported(&TIME_FORMAT_FIELD)) mInfo += "field ";
	if (S_OK == mpGraphMediaSeeking->IsFormatSupported(&TIME_FORMAT_FRAME)) mInfo += "frame ";
	if (S_OK == mpGraphMediaSeeking->IsFormatSupported(&TIME_FORMAT_MEDIA_TIME)) mInfo += "mediatime ";
	if (S_OK == mpGraphMediaSeeking->IsFormatSupported(&TIME_FORMAT_SAMPLE)) mInfo += "sample ";
	if (mInfo.back() == ' ')
		mInfo.pop_back();
	mInfo.push_back('\n');

	hr = mpSourcePin->ConnectedTo(~vdrefptr<IPin>());
	VDASSERT(SUCCEEDED(hr));

	hr = mpSampleGrabber->GetInputPin()->ConnectedTo(~vdrefptr<IPin>());
	VDASSERT(SUCCEEDED(hr));

	hr = mpSampleGrabber->GetOutputPin()->ConnectedTo(~vdrefptr<IPin>());
	VDASSERT(SUCCEEDED(hr));

	hr = mpGraph->Disconnect(mpSourcePin);
	if (FAILED(hr))
		throw MyError("Unable to disconnect source pin: hr=%08x", hr);

	hr = mpGraph->Disconnect(mpSampleGrabber->GetInputPin());
	if (FAILED(hr))
		throw MyError("Unable to disconnect sample grabber input pin: hr=%08x", hr);

	hr = mpSourcePin->ConnectedTo(~vdrefptr<IPin>());
	VDASSERT(FAILED(hr));

	hr = mpSampleGrabber->GetInputPin()->ConnectedTo(~vdrefptr<IPin>());
	VDASSERT(FAILED(hr));

	vdrefptr<IEnumMediaTypes> pMediaTypes;
	hr = mpSourcePin->EnumMediaTypes(~pMediaTypes);
	if (FAILED(hr))
		throw MyError("Unable to enumerate media types: hr=%08x", hr);

	AM_MEDIA_TYPE *amtRaw;
	ULONG fetched;
	REFERENCE_TIME period;
	bool found = false;

	while(S_OK == pMediaTypes->Next(1, &amtRaw, &fetched)) {
		MediaTypePtr amt;

		amt.Attach(amtRaw);

		if (amt->majortype == MEDIATYPE_Video) {
			bool valid = false;

			if (amt->formattype == FORMAT_VideoInfo) {
				const VIDEOINFOHEADER& vih = *(const VIDEOINFOHEADER *)amt->pbFormat;

				if (vih.AvgTimePerFrame) {
					period = vih.AvgTimePerFrame;
					VDFraction fr = VDFraction::reduce64(10000000, vih.AvgTimePerFrame);
					mStreamInfo.mSampleRate.mNumerator		= fr.getHi();
					mStreamInfo.mSampleRate.mDenominator	= fr.getLo();
				} else {
					period = 10000000 / 30;
					mStreamInfo.mSampleRate.mNumerator		= 30;
					mStreamInfo.mSampleRate.mDenominator	= 1;
				}

				mStreamInfo.mPixelAspectRatio.mNumerator = 1;
				mStreamInfo.mPixelAspectRatio.mDenominator = 1;

				mVideoInfo.mWidth		= vih.bmiHeader.biWidth;
				mVideoInfo.mHeight		= abs(vih.bmiHeader.biHeight);

				mInfo.append_sprintf("  Format\tVideoInfo %ux%u, %.4f fps: "
					, vih.bmiHeader.biWidth
					, vih.bmiHeader.biHeight
					, vih.AvgTimePerFrame ? 10000000.0 / vih.AvgTimePerFrame : 0);

				AppendVideoFormatName(mInfo, amt);

				mInfo += '\n';

				valid = true;
			} else if (amt->formattype == FORMAT_VideoInfo2) {
				const VIDEOINFOHEADER2& vih = *(const VIDEOINFOHEADER2 *)amt->pbFormat;

				if (vih.AvgTimePerFrame) {
					period = vih.AvgTimePerFrame;
					VDFraction fr = VDFraction::reduce64(10000000, vih.AvgTimePerFrame);
					mStreamInfo.mSampleRate.mNumerator		= fr.getHi();
					mStreamInfo.mSampleRate.mDenominator	= fr.getLo();
				} else {
					period = 10000000 / 30;
					mStreamInfo.mSampleRate.mNumerator		= 30;
					mStreamInfo.mSampleRate.mDenominator	= 1;
				}

				VDFraction frameAspect = VDFraction(vih.dwPictAspectRatioX, vih.dwPictAspectRatioY).reduce();
				VDFraction pixelAspect = frameAspect * VDFraction(abs(vih.bmiHeader.biHeight), vih.bmiHeader.biWidth);

				mStreamInfo.mPixelAspectRatio.mNumerator	= pixelAspect.getHi();
				mStreamInfo.mPixelAspectRatio.mDenominator	= pixelAspect.getLo();

				mVideoInfo.mWidth		= vih.bmiHeader.biWidth;
				mVideoInfo.mHeight		= abs(vih.bmiHeader.biHeight);

				mInfo.append_sprintf("  Format\tVideoInfo2 %ux%u, %.4f fps, aspect %u:%u, "
					, vih.bmiHeader.biWidth
					, vih.bmiHeader.biHeight
					, vih.AvgTimePerFrame ? 10000000.0 / vih.AvgTimePerFrame : 0
					, frameAspect.getHi()
					, frameAspect.getLo()
					);

				AppendVideoFormatName(mInfo, amt);
				mInfo += '\n';

				valid = true;
			}

			if (valid) {
				if (amt->subtype == MEDIASUBTYPE_RGB555)	mFormatSupportMask |= (1 << nsVDXPixmap::kPixFormat_XRGB1555);
				if (amt->subtype == MEDIASUBTYPE_RGB565)	mFormatSupportMask |= (1 << nsVDXPixmap::kPixFormat_RGB565);
				if (amt->subtype == MEDIASUBTYPE_RGB24)		mFormatSupportMask |= (1 << nsVDXPixmap::kPixFormat_RGB888);
				if (amt->subtype == MEDIASUBTYPE_RGB32)		mFormatSupportMask |= (1 << nsVDXPixmap::kPixFormat_XRGB8888);

				// Block YUV formats that aren't integral numbers of pixels in their chroma
				// planes; these are inconsistent between decoders.
				if (amt->subtype == MEDIASUBTYPE_UYVY) {
					if (mVideoInfo.mWidth & 1)
						valid = false;
					else
						mFormatSupportMask |= (1 << nsVDXPixmap::kPixFormat_YUV422_UYVY);
				}

				if (amt->subtype == MEDIASUBTYPE_YUY2) {
					if (mVideoInfo.mWidth & 1)
						valid = false;
					else
						mFormatSupportMask |= (1 << nsVDXPixmap::kPixFormat_YUV422_YUYV);
				}

				if (amt->subtype == MEDIASUBTYPE_YV12) {
					if ((mVideoInfo.mWidth | mVideoInfo.mHeight) & 1)
						valid = false;
					else
						mFormatSupportMask |= (1 << nsVDXPixmap::kPixFormat_YUV420_Planar);
				}

				if (amt->subtype == MEDIASUBTYPE_YVU9) {
					if ((mVideoInfo.mWidth | mVideoInfo.mHeight) & 3)
						valid = false;
					else
						mFormatSupportMask |= (1 << nsVDXPixmap::kPixFormat_YUV410_Planar);
				}
			}

			if (valid)
				found = true;
		}
	}

	if (!found)
		throw MyError("Could not find compatible decompression format.");

	mStreamInfo.mSampleCount = VDRoundToInt64((double)dur / (double)period);
	mVideoInfo.mDecoderModel = VDXVideoSourceInfo::kDecoderModelDefaultIP;
	mVideoInfo.mFlags = VDXVideoSourceInfo::kFlagKeyframeOnly;
	mFrameBuffer.resize(mVideoInfo.mWidth * mVideoInfo.mHeight * 4);

	DSListFilters(mpGraph, mInfo);
}

DSVideoDecoder::~DSVideoDecoder() {
	try {
		DSEngineCommand cmd(&mCommandSignal);
		cmd.SetMethodCallback<DSVideoDecoder, &DSVideoDecoder::ShutdownLocal>(this);
		mpEngine->SendCommand(&cmd);
	} catch(const MyError&) {
	}
}

void DSVideoDecoder::ShutdownLocal(void *) {
	if (mpGraphMediaControl) {
		mpGraphMediaControl->Stop();

		OAFilterState fs;
		mpGraphMediaControl->GetState(INFINITE, &fs);
	}

	mpRendererPin = NULL;
	mpRenderer = NULL;
	mpSampleGrabber = NULL;
	mpSourcePin = NULL;
	mpGraphMediaSeeking = NULL;
	mpGraphMediaEvent = NULL;
	mpGraphMediaControl = NULL;
	mpGraphMediaFilter = NULL;

	if (mGraphRegistration) {
		DSRemoveFromRot(mGraphRegistration);
		mGraphRegistration = 0;
	}

	mpGraph = NULL;
}

const void * VDXAPIENTRY DSVideoDecoder::DecodeFrame(const void *inputBuffer, uint32 data_len, bool is_preroll, sint64 sampleNumber, sint64 targetFrame) {
	try {
		if (data_len && !is_preroll) {
			DSEngineCommand cmd(&mCommandSignal);
			VDASSERT(!mCommandSignal.check());
			sint64 frame = *(const sint64 *)inputBuffer;
			cmd.SetMethodCallback<DSVideoDecoder, &DSVideoDecoder::DecodeFrameLocal>(this);
			cmd.SetExtraData(&frame);
			mpEngine->SendCommand(&cmd);
		}
	} catch(const MyError& e) {
		mpContext->mpCallbacks->SetError("%s", e.gets());
		return NULL;
	}
	return mFrameBuffer.data();
}

void DSVideoDecoder::DecodeFrameLocal(void *pvFrame) {
	bool forceSeek = false;

	for(;;) {
		// change time on sample grabber and see if it can grab the frame from its queue
		sint64 frame = *(sint64 *)pvFrame;
		LONGLONG samplePos = VDRoundToInt64(((double)frame + 0.5) * (double)mStreamInfo.mSampleRate.mDenominator / (double)mStreamInfo.mSampleRate.mNumerator * 10000000.0);

		VDDEBUG("Video: requesting sample at time %llu\n", samplePos);
		mbSampleLate = false;

		DSSampleGrabberCheckResult result = mpSampleGrabber->CheckForSample(samplePos, samplePos + 1);
		if (result == kDSSampleGrabberCheckResult_Completed) {
			// sweet
			return;
		}

		// issue a seek
		HRESULT hr;
		OAFilterState fs;

		REFERENCE_TIME lastSeekPos = 0;		// we don't need to initialize this, but it avoids uninited garbage in log message
		if (result != kDSSampleGrabberCheckResult_Partial && (forceSeek || !mpSampleGrabber->GetEstimatedStreamPosition(&lastSeekPos) || samplePos < lastSeekPos || samplePos > lastSeekPos + 5000000)) {
			LONGLONG startPos = VDRoundToInt64((double)frame * (double)mStreamInfo.mSampleRate.mDenominator / (double)mStreamInfo.mSampleRate.mNumerator * 10000000.0) - mSeekFudgeOffset;
			LONGLONG endPos = VDRoundToInt64((double)(frame + 2) * (double)mStreamInfo.mSampleRate.mDenominator / (double)mStreamInfo.mSampleRate.mNumerator * 10000000.0);

			if (startPos < 0)
				startPos = 0;

			if (endPos <= startPos)
				endPos = startPos + 1;

			VDDEBUG("Video: missed sample at %d; last seen was %d, reseeking to %d\n", (int)samplePos, (int)lastSeekPos, (int)startPos);

			// lock down the sample grabber
			mpSampleGrabber->ResetOneShot(samplePos, samplePos + 1, mSeekFudgeOffset >= kSeekFudgeOffsetMax);

			// We need to set a stop position here, or else the MP4 splitter has a habit of not sending
			// a new segment or end of stream notification when seeking near the end, resulting in a
			// hang.
			hr = mpGraphMediaSeeking->SetPositions(&startPos, AM_SEEKING_AbsolutePositioning, mbUseEndPosHack ? &endPos : NULL, mbUseEndPosHack ? AM_SEEKING_AbsolutePositioning : AM_SEEKING_NoPositioning);
			if (FAILED(hr))
				throw MyError("Unable to seek DirectShow graph: error code %08x", hr);

			if (hr == S_FALSE) {
				VDDEBUG("Video: SetPositions() returned S_FALSE.\n");
			}
		} else {
			VDDEBUG("Video: missed sample at %d; streaming through from %d\n", (int)samplePos, (int)lastSeekPos);
			if (mpSampleGrabber->ResumeOneShot(samplePos, samplePos + 1))
				return;
		}

		// run the graph
		hr = mpGraphMediaControl->GetState(INFINITE, &fs);
		if (FAILED(hr))
			throw MyError("Unable to run DirectShow graph: error code %08x", hr);

		if (fs != State_Running) {
			hr = mpGraphMediaControl->Run();
			if (FAILED(hr))
				throw MyError("Unable to run DirectShow graph: error code %08x", hr);

			// wait for state transition to complete
			hr = mpGraphMediaControl->GetState(INFINITE, &fs);
			if (FAILED(hr))
				throw MyError("Unable to run DirectShow graph: error code %08x", hr);
		}

		// wait for sample
		mpSampleGrabber->WaitForSample();

		if (!mbSampleLate)
			break;

		if (!mSeekFudgeOffset)
			mSeekFudgeOffset = kSeekFudgeOffsetMin;

		if (mSeekFudgeOffset < kSeekFudgeOffsetMax)
			mSeekFudgeOffset += mSeekFudgeOffset;

		forceSeek = true;
	}
}

bool DSVideoDecoder::Process(IMediaSample *pSample, const REFERENCE_TIME& timeBase) {
	REFERENCE_TIME t1, t2;
	pSample->GetTime(&t1, &t2);
	VDDEBUG("Receiving video: [%d, %d]\n", (int)(t1 + timeBase), (int)(t2 + timeBase));

	BYTE *p;
	HRESULT hr = pSample->GetPointer(&p);

	if (SUCCEEDED(hr)) {
		size_t size = pSample->GetActualDataLength();
		if (size > mFrameBuffer.size() * sizeof(mFrameBuffer[0]))
			size = mFrameBuffer.size() * sizeof(mFrameBuffer[0]);

		memcpy(mFrameBuffer.data(), p, size);
		mbFrameBufferValid = true;
	}

	mbSampleLate = false;
	return false;
}

void DSVideoDecoder::OnSampleMissed() {
	mbSampleLate = true;
}

uint32 VDXAPIENTRY DSVideoDecoder::GetDecodePadding() {
	return 0;
}

void VDXAPIENTRY DSVideoDecoder::Reset() {
	mbFrameBufferValid = false;
}

bool VDXAPIENTRY DSVideoDecoder::IsFrameBufferValid() {
	return mbFrameBufferValid;
}

const VDXPixmap& VDXAPIENTRY DSVideoDecoder::GetFrameBuffer() {
	return mpxFrameBuffer;
}

bool VDXAPIENTRY DSVideoDecoder::SetTargetFormat(int format, bool useDIBAlignment) {
	static const uint8 kFormatPriority[]={
		nsVDXPixmap::kPixFormat_RGB888,
		nsVDXPixmap::kPixFormat_XRGB8888,
		nsVDXPixmap::kPixFormat_YUV422_UYVY,
		nsVDXPixmap::kPixFormat_YUV422_YUYV,
		nsVDXPixmap::kPixFormat_YUV420_Planar,
		nsVDXPixmap::kPixFormat_XRGB1555,
		nsVDXPixmap::kPixFormat_RGB565,
		nsVDXPixmap::kPixFormat_YUV410_Planar,
	};

	if (!format) {
		for(int i=0; i<sizeof(kFormatPriority)/sizeof(kFormatPriority[0]); ++i) {
			int j = kFormatPriority[i];
			if (mFormatSupportMask & (1 << j)) {
				format = j;
				break;
			}
		}
	}

	if (format && format == mpxFrameBuffer.format)
		return true;

	if (format >= 32 || !(mFormatSupportMask & (1<<format)))
		return false;

	mbFrameBufferValid = false;

	try {
		DSEngineCommand cmd(&mCommandSignal);
		cmd.SetMethodCallback<DSVideoDecoder, &DSVideoDecoder::SetFormatLocal>(this);
		cmd.SetExtraData(&format);
		mpEngine->SendCommand(&cmd);
	} catch(const MyError&) {
		return false;
	}

	return true;
}

void DSVideoDecoder::SetFormatLocal(void *pvFormat) {
	int format = *(const int *)pvFormat;

	// stop the graph
	HRESULT hr = mpGraphMediaControl->Stop();
	if (FAILED(hr))
		throw MyError("Unable to stop DirectShow graph: error code %08x", hr);

	// wait for stop to complete
	OAFilterState fs;
	hr = mpGraphMediaControl->GetState(INFINITE, &fs);
	if (FAILED(hr))
		throw MyError("Unable to stop DirectShow graph: error code %08x", hr);

	hr = mpGraph->Disconnect(mpSourcePin);
	if (FAILED(hr))
		throw MyError("Unable to disconnect source pin: hr=%08x", hr);

	hr = mpGraph->Disconnect(mpSampleGrabber->GetInputPin());
	if (FAILED(hr))
		throw MyError("Unable to disconnect sample grabber input pin: hr=%08x", hr);

	hr = mpGraph->Disconnect(mpSampleGrabber->GetOutputPin());
	if (FAILED(hr))
		throw MyError("Unable to disconnect sample grabber output pin: hr=%08x", hr);

	hr = mpGraph->Disconnect(mpRendererPin);
	if (FAILED(hr))
		throw MyError("Unable to disconnect renderer pin: hr=%08x", hr);

	char *fbbase = (char *)mFrameBuffer.data();
	mpxFrameBuffer.w		= mVideoInfo.mWidth;
	mpxFrameBuffer.h		= mVideoInfo.mHeight;
	mpxFrameBuffer.format	= format;
	mpxFrameBuffer.data		= fbbase;
	mpxFrameBuffer.palette	= NULL;
	mpxFrameBuffer.data2	= NULL;
	mpxFrameBuffer.pitch2	= 0;
	mpxFrameBuffer.data3	= NULL;
	mpxFrameBuffer.pitch3	= 0;

	// Rounding/truncation is omitted for sizes below because we block such formats as ill-defined.

	bool verticallyFlippable = false;

	switch(format) {
		case nsVDXPixmap::kPixFormat_XRGB1555:
			mpxFrameBuffer.pitch	= -((mpxFrameBuffer.w*2 + 3) & ~3);
			mpxFrameBuffer.data		= fbbase - mpxFrameBuffer.pitch * (mpxFrameBuffer.h - 1);
			mpSampleGrabber->SetMediaSubType(&MEDIASUBTYPE_RGB555);
			verticallyFlippable = true;
			break;
		case nsVDXPixmap::kPixFormat_RGB565:
			mpxFrameBuffer.pitch	= -((mpxFrameBuffer.w*2 + 3) & ~3);
			mpxFrameBuffer.data		= fbbase - mpxFrameBuffer.pitch * (mpxFrameBuffer.h - 1);
			mpSampleGrabber->SetMediaSubType(&MEDIASUBTYPE_RGB565);
			verticallyFlippable = true;
			break;
		case nsVDXPixmap::kPixFormat_RGB888:
			mpxFrameBuffer.pitch	= -((mpxFrameBuffer.w*3 + 3) & ~3);
			mpxFrameBuffer.data		= fbbase - mpxFrameBuffer.pitch * (mpxFrameBuffer.h - 1);
			mpSampleGrabber->SetMediaSubType(&MEDIASUBTYPE_RGB24);
			verticallyFlippable = true;
			break;
		case nsVDXPixmap::kPixFormat_XRGB8888:
			mpxFrameBuffer.pitch	= -mpxFrameBuffer.w * 4;
			mpxFrameBuffer.data		= fbbase - mpxFrameBuffer.pitch * (mpxFrameBuffer.h - 1);
			mpSampleGrabber->SetMediaSubType(&MEDIASUBTYPE_RGB32);
			verticallyFlippable = true;
			break;
		case nsVDXPixmap::kPixFormat_YUV422_UYVY:
			mpxFrameBuffer.pitch	= mpxFrameBuffer.w * 2;
			mpSampleGrabber->SetMediaSubType(&MEDIASUBTYPE_UYVY);
			break;
		case nsVDXPixmap::kPixFormat_YUV422_YUYV:
			mpxFrameBuffer.pitch	= mpxFrameBuffer.w * 2;
			mpSampleGrabber->SetMediaSubType(&MEDIASUBTYPE_YUY2);
			break;
		case nsVDXPixmap::kPixFormat_YUV420_Planar:
			mpxFrameBuffer.pitch	= mpxFrameBuffer.w;
			mpxFrameBuffer.data3	= fbbase + mpxFrameBuffer.h * mpxFrameBuffer.pitch;
			mpxFrameBuffer.pitch3	= mpxFrameBuffer.w >> 1;
			mpxFrameBuffer.data2	= (char *)mpxFrameBuffer.data3 + (mpxFrameBuffer.h >> 1) * mpxFrameBuffer.pitch3;
			mpxFrameBuffer.pitch2	= mpxFrameBuffer.pitch3;
			mpSampleGrabber->SetMediaSubType(&MEDIASUBTYPE_YV12);
			break;
		case nsVDXPixmap::kPixFormat_YUV410_Planar:
			mpxFrameBuffer.pitch	= mpxFrameBuffer.w;
			mpxFrameBuffer.data3	= fbbase + mpxFrameBuffer.h * mpxFrameBuffer.pitch;
			mpxFrameBuffer.pitch3	= mpxFrameBuffer.w >> 2;
			mpxFrameBuffer.data2	= (char *)mpxFrameBuffer.data3 + (mpxFrameBuffer.h >> 2) * mpxFrameBuffer.pitch3;
			mpxFrameBuffer.pitch2	= mpxFrameBuffer.pitch3;
			mpSampleGrabber->SetMediaSubType(&MEDIASUBTYPE_YVU9);
			break;
	}	

	hr = mpGraph->ConnectDirect(mpSourcePin, mpSampleGrabber->GetInputPin(), NULL);
	if (FAILED(hr))
		throw MyError("Unable to connect sample grabber to source: hr=%08x", hr);

	if (verticallyFlippable) {
		const AM_MEDIA_TYPE *mtype = mpSampleGrabber->GetInputPin()->GetConnectedMediaType();
		if (mtype) {
			bool verticallyFlipped = false;

			if (mtype->formattype == FORMAT_VideoInfo) {
				const VIDEOINFOHEADER *vih = (const VIDEOINFOHEADER *)mtype->pbFormat;

				if (vih->bmiHeader.biHeight < 0)
					verticallyFlipped = true;
			} else if (mtype->formattype == FORMAT_VideoInfo2) {
				const VIDEOINFOHEADER2 *vih2 = (const VIDEOINFOHEADER2 *)mtype->pbFormat;

				if (vih2->bmiHeader.biHeight < 0)
					verticallyFlipped = true;
			}

			if (verticallyFlipped) {
				mpxFrameBuffer.data = (char *)mpxFrameBuffer.data + mpxFrameBuffer.pitch * (mpxFrameBuffer.h - 1);
				mpxFrameBuffer.pitch = -mpxFrameBuffer.pitch;
			}
		}
	}

	hr = mpGraph->ConnectDirect(mpSampleGrabber->GetOutputPin(), mpRendererPin, NULL);
	if (FAILED(hr))
		throw MyError("Unable to connect sample grabber to renderer: hr=%08x", hr);
}

bool VDXAPIENTRY DSVideoDecoder::SetDecompressedFormat(const VDXBITMAPINFOHEADER *pbih) {
	return false;
}

bool VDXAPIENTRY DSVideoDecoder::IsDecodable(sint64 sample_num) {
	return true;
}

const void * VDXAPIENTRY DSVideoDecoder::GetFrameBufferBase() {
	return mFrameBuffer.data();
}

///////////////////////////////////////////////////////////////////////////////

DSVideoSource::DSVideoSource(const VDXInputDriverContext *context, IDSEngine *engine, const wchar_t *fn)
	: mpEngine(engine)
	, mpContext(context)
	, mFilename(fn)
{
	mpCachedDecoder = new DSVideoDecoder(mpContext, mpEngine, mFilename.c_str());
	mInfo = mpCachedDecoder->GetInfo();

	mStreamInfo		= mpCachedDecoder->GetStreamInfo();
	mVideoInfo		= mpCachedDecoder->GetVideoInfo();
}

DSVideoSource::~DSVideoSource() {
}

void DSVideoSource::GetInfo(VDStringA& s) {
	s += mInfo;
}

int		VDXAPIENTRY DSVideoSource::AddRef() {
	return vdxunknown<IVDXVideoSource>::AddRef();
}

int		VDXAPIENTRY DSVideoSource::Release() {
	return vdxunknown<IVDXVideoSource>::Release();
}

void *	VDXAPIENTRY DSVideoSource::AsInterface(uint32 id) {
	switch(id) {
		case IVDXStreamSource::kIID:
			return static_cast<IVDXStreamSource *>(this);

		case IVDXStreamSourceV3::kIID:
			return static_cast<IVDXStreamSourceV3 *>(this);
	}

	return vdxunknown<IVDXVideoSource>::AsInterface(id);
}

void VDXAPIENTRY DSVideoSource::GetStreamSourceInfo(VDXStreamSourceInfo& info) {
	info = mStreamInfo;
}

bool VDXAPIENTRY DSVideoSource::Read(sint64 lStart, uint32 lCount, void *lpBuffer, uint32 cbBuffer, uint32 *lBytesRead, uint32 *lSamplesRead) {
	if (!lpBuffer || cbBuffer < sizeof(sint64)) {
		*lBytesRead = 8;
		*lSamplesRead = 1;
		return !lpBuffer;
	}

	if (!lCount) {
		*lBytesRead = 0;
		*lSamplesRead = 0;
		return true;
	}

	*lSamplesRead = 1;
	*lBytesRead = sizeof(sint64);

	*(sint64 *)lpBuffer = lStart;
	return true;
}

const void *VDXAPIENTRY DSVideoSource::GetDirectFormat() {
	return NULL;
}

int	VDXAPIENTRY DSVideoSource::GetDirectFormatLen() {
	return 0;
}

IVDXStreamSource::ErrorMode	VDXAPIENTRY DSVideoSource::GetDecodeErrorMode() {
	return kErrorModeReportAll;
}

void VDXAPIENTRY DSVideoSource::SetDecodeErrorMode(ErrorMode mode) {
}

bool VDXAPIENTRY DSVideoSource::IsDecodeErrorModeSupported(ErrorMode mode) {
	return mode == kErrorModeReportAll;
}

bool VDXAPIENTRY DSVideoSource::IsVBR() {
	return false;
}

sint64 VDXAPIENTRY DSVideoSource::TimeToPositionVBR(sint64 us) {
	return -1;
}

sint64 VDXAPIENTRY DSVideoSource::PositionToTimeVBR(sint64 samples) {
	return -1;
}

void VDXAPIENTRY DSVideoSource::GetStreamSourceInfoV3(VDXStreamSourceInfoV3& infov3) {
	infov3.mfccHandler = 0;
	infov3.mFlags = 0;
	
	GetStreamSourceInfo(infov3.mInfo);
}

void VDXAPIENTRY DSVideoSource::GetVideoSourceInfo(VDXVideoSourceInfo& info) {
	info = mVideoInfo;
}

bool VDXAPIENTRY DSVideoSource::CreateVideoDecoderModel(IVDXVideoDecoderModel **) {
	return false;
}

bool VDXAPIENTRY DSVideoSource::CreateVideoDecoder(IVDXVideoDecoder **ppDecoder) {
	if (mpCachedDecoder) {
		*ppDecoder = mpCachedDecoder.release();
		return true;
	}

	*ppDecoder = new DSVideoDecoder(mpContext, mpEngine, mFilename.c_str());
	if (!*ppDecoder) {
		mpContext->mpCallbacks->SetErrorOutOfMemory();
		return false;
	}

	(*ppDecoder)->AddRef();
	return true;
}

void VDXAPIENTRY DSVideoSource::GetSampleInfo(sint64 sample_num, VDXVideoFrameInfo& frameInfo) {
	frameInfo.mBytePosition = -1;
	frameInfo.mFrameType = kVDXVFT_Independent;
	frameInfo.mTypeChar = 'K';
}

bool VDXAPIENTRY DSVideoSource::IsKey(sint64 sample_num) {
	return true;
}

sint64 VDXAPIENTRY DSVideoSource::GetFrameNumberForSample(sint64 sample_num) {
	return sample_num;
}

sint64 VDXAPIENTRY DSVideoSource::GetSampleNumberForFrame(sint64 frame_num) {
	return frame_num;
}

sint64 VDXAPIENTRY DSVideoSource::GetRealFrame(sint64 frame_num) {
	return frame_num;
}

sint64 VDXAPIENTRY DSVideoSource::GetSampleBytePosition(sint64 sample_num) {
	return -1;
}
